from argparse import ArgumentParser
from functools import partial
import json
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterable, Mapping

from typepigeon import convert_value

from coupledmodeldriver.configure import ModelJSON
from coupledmodeldriver.generate.adcirc.base import ADCIRCJSON
from coupledmodeldriver.generate.adcirc.check import (
    check_adcirc_completion,
    CompletionStatus,
    is_adcirc_run_directory,
)
from coupledmodeldriver.generate.schism.base import SCHISMJSON
from coupledmodeldriver.generate.schism.check import (
    check_schism_completion,
    CompletionStatus,
    is_schism_run_directory,
)
from coupledmodeldriver.utilities import ProcessPoolExecutorStackTraced

MODELS = {model.name.lower(): model for model in ModelJSON.__subclasses__()}


def parse_check_completion_arguments():
    argument_parser = ArgumentParser()
    argument_parser.add_argument(
        'directory',
        nargs='*',
        default=Path.cwd(),
        help='directory containing model run configuration',
    )
    argument_parser.add_argument('--model', help='model that is running, one of: `ADCIRC`')
    argument_parser.add_argument(
        '--verbose', action='store_true', help='list all errors and problems with runs'
    )

    arguments = argument_parser.parse_args()

    model = arguments.model
    if model is not None:
        model = MODELS[model.lower()]

    directory = convert_value(arguments.directory, [Path])
    if len(directory) == 1:
        directory = directory[0]

    return {
        'directory': directory,
        'model': model,
        'verbose': arguments.verbose,
    }


def is_model_directory(directory: PathLike, model: ModelJSON = None) -> bool:
    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if model is None:
        model = ADCIRCJSON

    if model == ADCIRCJSON:
        is_model_directory = is_adcirc_run_directory(directory)
    elif model == SCHISMJSON:
        is_model_directory = is_schism_run_directory(directory)
    else:
        raise NotImplementedError(f'model "{model}" not implemented')

    return is_model_directory


def check_model_directory(
    directory: PathLike, verbose: bool = False, model: ModelJSON = None
) -> Dict[str, Any]:
    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    if model is None:
        model = ADCIRCJSON

    if model == ADCIRCJSON:
        return check_adcirc_completion(directory, verbose=verbose)
    elif model == SCHISMJSON:
        return check_schism_completion(directory, verbose=verbose)
    else:
        raise NotImplementedError(f'model "{model}" not implemented')


def check_completion(
    directory: PathLike = None, verbose: bool = False, model: ModelJSON = None
) -> Dict[str, Any]:
    """
    check the completion status of a running model

    :param directory: directory containing model run configuration
    :param verbose: list all errors and problems with runs
    :param model: model that is running, one of: ``ADCIRC``
    :return: JSON output of completion status
    """

    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path) and (
        not isinstance(directory, Iterable) or isinstance(directory, str)
    ):
        directory = Path(directory)

    completion_status = {}
    if isinstance(directory, Iterable):
        with ProcessPoolExecutorStackTraced() as process_pool:
            subdirectory_completion_statuses = process_pool.map(
                partial(check_completion, model=model, verbose=verbose),
                (subdirectory for subdirectory in directory if Path(subdirectory).is_dir()),
            )
            for subdirectory_completion_status in subdirectory_completion_statuses:
                completion_status.update(subdirectory_completion_status)
    elif isinstance(directory, Path):
        if model is None:
            # model = ADCIRCJSON
            for model_type in MODELS.values():
                if not is_model_directory(directory, model=model_type):
                    continue
                model = model_type
                break

        if is_model_directory(directory, model=model):
            completion = check_model_directory(
                directory=directory, verbose=verbose, model=model
            )
            for key in [
                key
                for key, value in completion.items()
                if isinstance(value, Mapping) and len(value) == 0
            ]:
                del completion[key]
            completion['status'] = completion['status'].name.lower()
            completion['progress'] = f'{completion["progress"]}%'
            completion_status[directory.name] = completion
        else:
            subdirectory_completion_statuses = check_completion(
                directory=directory.iterdir(), verbose=verbose, model=model
            )
            if len(subdirectory_completion_statuses) > 0:
                completion_status[directory.name] = subdirectory_completion_statuses

    try:
        # sort by progress percentage (reversed), then completion status, then by run number, then alphabetically
        completion_status = dict(
            sorted(
                completion_status.items(),
                key=lambda item: (
                    -float(
                        (
                            item[1]['progress']
                            if isinstance(item[1], Mapping)
                            else item[1].split(' - ')[1]
                        )[:-1]
                    ),
                    CompletionStatus[
                        (
                            item[1]['status']
                            if isinstance(item[1], Mapping)
                            else item[1].split(' - ')[0]
                        ).upper()
                    ].value,
                    float(item[0].split('_')[-1])
                    if '_' in item[0] and item[0].split('_')[-1].isdecimal()
                    else -1,
                    item[0],
                ),
            )
        )
    except KeyError:
        pass

    if not verbose:
        for key, value in completion_status.items():
            if isinstance(value, Mapping) and 'status' in value and 'progress' in value:
                completion_status[key] = f'{value["status"]} - {value["progress"]}'

    if (
        isinstance(completion_status, Mapping)
        and len(completion_status) > 0
        and isinstance(directory, Path)
    ):
        directory_completion_status = completion_status[directory.name]
        if (
            isinstance(directory_completion_status, Mapping)
            and len(directory_completion_status) > 0
        ):
            # collapse statuses if all statuses in the directory are the same
            statuses = list(directory_completion_status.values())
            if all(status == statuses[0] for status in statuses):
                completion_status[directory.name] = statuses[0]

    return completion_status


def main():
    completion_status = check_completion(**parse_check_completion_arguments())
    print(json.dumps(completion_status, indent=4))


if __name__ == '__main__':
    main()



from argparse import ArgumentParser
from pathlib import Path

from typepigeon import convert_value

from coupledmodeldriver.generate import generate_schism_configuration


def parse_generate_schism_arguments():
    argument_parser = ArgumentParser()

    argument_parser.add_argument(
        '--configuration-directory',
        default=Path().cwd(),
        help='path containing JSON configuration files',
    )
    argument_parser.add_argument(
        '--output-directory', default=None, help='path to store generated configuration files'
    )
    argument_parser.add_argument(
        '--relative-paths',
        action='store_true',
        help='use relative paths in output configuration',
    )
    argument_parser.add_argument(
        '--overwrite', action='store_true', help='overwrite existing files',
    )
    argument_parser.add_argument(
        '--parallel', action='store_true', help='generate configurations concurrently',
    )

    arguments = argument_parser.parse_args()

    return {
        'configuration_directory': convert_value(arguments.configuration_directory, Path),
        'output_directory': convert_value(arguments.output_directory, Path),
        'relative_paths': arguments.relative_paths,
        'overwrite': arguments.overwrite,
        'parallel': arguments.parallel,
    }


def main():
    generate_schism_configuration(**parse_generate_schism_arguments())


if __name__ == '__main__':
    main()



from abc import ABC
from copy import copy
from os import PathLike
from pathlib import Path
from typing import Any, Collection, Dict, List, Mapping, Set, Union

from coupledmodeldriver.configure.base import ConfigurationJSON, ModelDriverJSON
from coupledmodeldriver.configure.forcings.base import (
    ADCIRCPY_FORCING_CLASSES,
    PYSCHISM_FORCING_CLASSES,
    ForcingJSON,
)
from coupledmodeldriver.utilities import LOGGER


class RunConfiguration(ABC):
    """
    abstraction of a set of model run configurations, encapsulated in JSON files
    """

    REQUIRED: Set[ConfigurationJSON] = {ModelDriverJSON}
    SUPPLEMENTARY: Set[ConfigurationJSON] = set()

    def __init__(self, configurations: List[ConfigurationJSON]):
        self.__configurations = {}
        self.configurations = configurations

    def perturb(self, relative_to: PathLike = None) -> Dict[str, 'RunConfiguration']:
        perturbed_configurations = {}
        if 'modeldriver' in self:
            perturbations = self['modeldriver']['perturbations']

            for run, run_perturbations in perturbations.items():
                instance = copy(self) if relative_to is None else self.relative_to(relative_to)
                if run_perturbations is not None and len(run_perturbations) > 0:
                    for name, configuration_perturbations in run_perturbations.items():
                        if name in instance:
                            instance[name].update(configuration_perturbations)
                        else:
                            LOGGER.warning(f'configuration "{name}" not found to perturb')
                perturbed_configurations[run] = instance
        else:
            perturbed_configurations = {}

        return perturbed_configurations

    @property
    def configurations(self) -> List[ConfigurationJSON]:
        return list(self.__configurations.values())

    @configurations.setter
    def configurations(self, configurations: List[ConfigurationJSON]):
        if isinstance(configurations, Collection) and not isinstance(configurations, Mapping):
            configurations = {
                entry.name.lower(): entry
                for entry in configurations
                if isinstance(entry, ConfigurationJSON)
            }
        for name, configuration in configurations.items():
            self[name] = configuration

    def move_paths(self, relative: PathLike):
        for configuration in self.configurations:
            configuration.move_paths(relative)

    def relative_to(self, path: PathLike, inplace: bool = False) -> 'RunConfiguration':
        instance = copy(self) if not inplace else self
        for configuration in instance.configurations:
            configuration.relative_to(path, inplace=True)
        return instance

    def __contains__(self, configuration: Union[str, ConfigurationJSON]) -> bool:
        if isinstance(configuration, str):
            return configuration.lower() in self.__configurations
        else:
            if not isinstance(configuration, ConfigurationJSON):
                configuration = from_user_input(configuration)
            return configuration in self.__configurations.values()

    def __getitem__(self, name: str) -> ConfigurationJSON:
        return self.__configurations[name.lower()]

    def __setitem__(self, name: str, configuration: ConfigurationJSON):
        if not isinstance(configuration, ConfigurationJSON):
            configuration = from_user_input(configuration)
        self.__configurations[name.lower()] = configuration

    def __iter__(self) -> ConfigurationJSON:
        yield from self.__configurations

    def items(self) -> (str, ConfigurationJSON):
        yield from self.__configurations.items()

    def __copy__(self) -> 'RunConfiguration':
        return self.__class__([copy(configuration) for configuration in self.configurations])

    def add(self, configuration: ConfigurationJSON) -> str:
        if not isinstance(configuration, ConfigurationJSON):
            configuration = from_user_input(configuration)
        self.__configurations[configuration.name.lower()] = configuration
        return configuration.name.lower()

    @classmethod
    def read_directory(
        cls, directory: PathLike, required: List[type] = None, supplementary: List[type] = None
    ) -> 'RunConfiguration':
        if not isinstance(directory, Path):
            directory = Path(directory)
        if directory.is_file():
            directory = directory.parent
        if required is None:
            required = set()
        required.update(RunConfiguration.REQUIRED)
        if supplementary is None:
            supplementary = set()
        supplementary.update(RunConfiguration.SUPPLEMENTARY)

        configurations = []
        for configuration_class in required:
            filename = directory / configuration_class.default_filename
            if filename.exists():
                configurations.append(configuration_class.from_file(filename))
            else:
                raise FileNotFoundError(f'missing required configuration file "{filename}"')

        for configuration_class in supplementary:
            filename = directory / configuration_class.default_filename
            if filename.exists():
                configurations.append(configuration_class.from_file(filename))

        return cls.from_configurations(configurations)

    def write_directory(
        self, directory: PathLike, absolute: bool = False, overwrite: bool = False
    ):
        """
        :param directory: directory in which to write generated JSON configuration files
        :param absolute: whether to write absolute paths
        :param overwrite: whether to overwrite existing files
        """

        if not isinstance(directory, Path):
            directory = Path(directory)

        if not directory.exists():
            directory.mkdir(parents=True, exist_ok=True)

        for configuration in self.__configurations.values():
            configuration.to_file(directory, absolute=absolute, overwrite=overwrite)

    @classmethod
    def from_configurations(
        cls, configurations: List[ConfigurationJSON]
    ) -> 'RunConfiguration':
        return cls(configurations)


def from_user_input(value: Any) -> ConfigurationJSON:
    """
    construct a configuration JSON object from a forcing object, a file, or a string

    :param value: value to parse; either an ADCIRCpy forcing object, file path, or a JSON string value
    :return: configuration JSON object
    """

    if isinstance(value, str):
        if Path(value).exists():
            value = ConfigurationJSON.from_file(value)
        else:
            value = ConfigurationJSON.from_string(value)
    elif isinstance(value, ADCIRCPY_FORCING_CLASSES):
        value = ForcingJSON.from_adcircpy(value)
    elif isinstance(value, PYSCHISM_FORCING_CLASSES):
        value = ForcingJSON.from_pyschism(value)
    return value



import concurrent.futures
from copy import copy, deepcopy
from datetime import datetime, timedelta
from enum import Enum
import os
from os import PathLike
from pathlib import Path
from typing import Union

from adcircpy import AdcircMesh
from nemspy import ModelingSystem

from coupledmodeldriver import Platform
from coupledmodeldriver.generate.adcirc.configure import (
    ADCIRCRunConfiguration,
    NEMSADCIRCRunConfiguration,
)
from coupledmodeldriver.generate.adcirc.script import (
    AdcircRunJob,
    AdcircSetupJob,
    AswipCommand,
    AdcircEnsembleRunScript,
    AdcircEnsembleCleanupScript,
)
from coupledmodeldriver.script import SlurmEmailType
from coupledmodeldriver.utilities import (
    create_symlink,
    get_logger,
    ProcessPoolExecutorStackTraced,
)

LOGGER = get_logger('cplmdldrv')


class RunPhase(Enum):
    """
    options for run phase of ADCIRC execuction
    """

    COLDSTART = 'coldstart'
    HOTSTART = 'hotstart'


def generate_adcirc_configuration(
    configuration_directory: PathLike,
    output_directory: PathLike = None,
    relative_paths: bool = False,
    overwrite: bool = False,
    parallel: bool = True,
):
    """
    generate an ADCIRC run configuration from JSON configuration files generated by ``initialize_adcirc``

    :param configuration_directory: path containing JSON configuration files
    :param output_directory: path to store generated configuration files
    :param relative_paths: whether to write relative paths in generated configuration files
    :param overwrite: whether to overwrite existing files
    :param parallel: generate configurations concurrently
    """

    start_time = datetime.now()

    if not isinstance(configuration_directory, Path):
        configuration_directory = Path(configuration_directory)

    if output_directory is None:
        output_directory = configuration_directory
    elif not isinstance(output_directory, Path):
        output_directory = Path(output_directory)

    output_directory = output_directory.resolve()
    if not output_directory.is_absolute():
        output_directory = output_directory.absolute()

    if not output_directory.exists():
        os.makedirs(output_directory, exist_ok=True)

    if configuration_directory.absolute().resolve() != Path.cwd():
        starting_directory = Path.cwd()
        os.chdir(configuration_directory)
        configuration_directory = Path.cwd()
    else:
        starting_directory = None

    use_nems = 'configure_nems.json' in [
        filename.name.lower() for filename in configuration_directory.iterdir()
    ]

    if use_nems:
        LOGGER.debug(f'generating NEMS configuration')
        base_configuration = NEMSADCIRCRunConfiguration.read_directory(configuration_directory)
    else:
        LOGGER.debug(f'generating ADCIRC-only configuration')
        base_configuration = ADCIRCRunConfiguration.read_directory(configuration_directory)

    base_configuration.move_paths(configuration_directory)

    platform = base_configuration['modeldriver']['platform']

    job_duration = base_configuration['slurm']['job_duration']
    slurm_account = base_configuration['slurm']['account']
    partition = base_configuration['slurm']['partition']
    email_type = base_configuration['slurm']['email_type']
    email_address = base_configuration['slurm']['email_address']

    original_fort13_filename = base_configuration['adcirc']['fort_13_path']
    original_fort14_filename = base_configuration['adcirc']['fort_14_path']
    adcirc_processors = base_configuration['adcirc']['processors']
    spinup_duration = base_configuration['adcirc']['tidal_spinup_duration']
    use_original_mesh = base_configuration['adcirc']['use_original_mesh']

    if original_fort14_filename is None or not original_fort14_filename.exists():
        raise FileNotFoundError(f'mesh XY not found at "{original_fort14_filename}"')

    local_fort13_filename = output_directory / 'fort.13'
    local_fort14_filename = output_directory / 'fort.14'
    local_fort15_filename = output_directory / 'fort.15'

    do_spinup = spinup_duration is not None

    run_phase = 'HOTSTART' if do_spinup else 'COLDSTART'

    if slurm_account is None:
        slurm_account = platform.value['slurm_account']

    ensemble_run_script_filename = output_directory / f'run_{platform.name.lower()}.sh'
    ensemble_cleanup_script_filename = output_directory / f'cleanup.sh'

    if use_original_mesh:
        LOGGER.info(
            f'using original mesh from "{os.path.relpath(original_fort14_filename.resolve(), Path.cwd())}"'
        )
        if original_fort13_filename.exists():
            create_symlink(original_fort13_filename, local_fort13_filename)
        create_symlink(original_fort14_filename, local_fort14_filename)
    elif overwrite or not local_fort14_filename.exists():
        LOGGER.info(
            f'rewriting original mesh to "{os.path.relpath(local_fort14_filename.resolve(), Path.cwd())}"'
        )
        adcircpy_driver = base_configuration.adcircpy_driver
        try:
            adcircpy_driver.write(
                output_directory,
                overwrite=overwrite,
                fort13='fort.13' if original_fort13_filename.exists() else None,
                fort14='fort.14',
                fort15='fort.15',
                fort22=None,
                coldstart=None,
                hotstart=None,
                driver=None,
            )
        except Exception as error:
            LOGGER.warning(error)

    if local_fort15_filename.exists():
        os.remove(local_fort15_filename)

    if local_fort13_filename.exists():
        base_configuration['adcirc']['fort_13_path'] = local_fort13_filename
    if local_fort14_filename.exists():
        base_configuration['adcirc']['fort_14_path'] = local_fort14_filename

    runs_directory = output_directory / 'runs'
    if not runs_directory.exists():
        runs_directory.mkdir(parents=True, exist_ok=True)

    perturbations = base_configuration.perturb()

    LOGGER.info(
        f'generating {len(perturbations)} run configuration(s) in "{os.path.relpath(runs_directory.resolve(), Path.cwd())}"'
    )

    if parallel:
        process_pool = ProcessPoolExecutorStackTraced()
        LOGGER.info(f'leveraging {os.cpu_count()} processor(s)')
    else:
        process_pool = None

    futures = []

    if do_spinup:
        spinup_directory = output_directory / 'spinup'

        spinup_configuration = copy(base_configuration)

        if not spinup_configuration.files_exist(spinup_directory):
            spinup_kwargs = {
                'directory': spinup_directory,
                'configuration': spinup_configuration,
                'duration': spinup_duration,
                'local_fort13_filename': local_fort13_filename,
                'local_fort14_filename': local_fort14_filename,
                'link_mesh': True,
                'relative_paths': relative_paths,
                'overwrite': overwrite,
                'platform': platform,
                'adcirc_processors': adcirc_processors,
                'slurm_account': slurm_account,
                'job_duration': job_duration,
                'partition': partition,
                'email_type': email_type,
                'email_address': email_address,
            }

            if parallel:
                futures.append(process_pool.submit(write_spinup_directory, **spinup_kwargs))
            else:
                spinup_directory = write_spinup_directory(**spinup_kwargs)
                LOGGER.info(f'wrote configuration to "{spinup_directory}"')
    else:
        spinup_directory = None

    for run_name, run_configuration in perturbations.items():
        run_directory = runs_directory / run_name

        if not run_configuration.files_exist(run_directory):
            run_kwargs = {
                'directory': run_directory,
                'name': run_name,
                'phase': run_phase,
                'configuration': run_configuration,
                'local_fort13_filename': local_fort13_filename,
                'local_fort14_filename': local_fort14_filename,
                'link_mesh': use_original_mesh,
                'relative_paths': relative_paths,
                'overwrite': overwrite,
                'platform': platform,
                'adcirc_processors': adcirc_processors,
                'slurm_account': slurm_account,
                'job_duration': job_duration,
                'partition': partition,
                'email_type': email_type,
                'email_address': email_address,
                'do_spinup': do_spinup,
                'spinup_directory': spinup_directory,
            }

            if parallel:
                # destroy stored copy of adcircpy mesh, because it cannot be pickled across processes
                run_configuration.adcircpy_mesh = None
                futures.append(process_pool.submit(write_run_directory, **run_kwargs))
            else:
                write_run_directory(**run_kwargs)
                LOGGER.info(f'wrote configuration to "{run_directory}"')

    if parallel:
        for completed_future in concurrent.futures.as_completed(futures):
            LOGGER.info(f'wrote configuration to "{completed_future.result()}"')

    cleanup_script = AdcircEnsembleCleanupScript()
    LOGGER.debug(
        f'writing cleanup script "{os.path.relpath(ensemble_cleanup_script_filename.resolve(), Path.cwd())}"'
    )
    cleanup_script.write(filename=ensemble_cleanup_script_filename, overwrite=overwrite)

    LOGGER.info(
        f'writing ensemble run script "{os.path.relpath(ensemble_run_script_filename.resolve(), Path.cwd())}"'
    )
    run_job_script = AdcircEnsembleRunScript(
        platform=platform,
        commands=[
            'echo deleting previous ADCIRC output',
            f'sh {ensemble_cleanup_script_filename.name}',
            'echo deleting previous ADCIRC logs',
            'rm spinup/*.log runs/*/*.log',
        ],
        run_spinup=do_spinup,
    )
    run_job_script.write(ensemble_run_script_filename, overwrite=overwrite)

    if starting_directory is not None:
        os.chdir(starting_directory)

    LOGGER.info(f'finished in {datetime.now() - start_time}')


def write_spinup_directory(
    directory: PathLike,
    configuration: Union[ADCIRCRunConfiguration, NEMSADCIRCRunConfiguration],
    duration: timedelta,
    local_fort14_filename: PathLike,
    local_fort13_filename: PathLike = None,
    link_mesh: bool = False,
    relative_paths: bool = False,
    overwrite: bool = False,
    platform: Platform = None,
    adcirc_processors: int = None,
    slurm_account: str = None,
    job_duration: timedelta = None,
    partition: str = None,
    email_type: SlurmEmailType = None,
    email_address: str = None,
) -> Path:
    """
    write an ADCIRC tidal spinup configuration to the given directory

    :param directory: path to output directory
    :param configuration: run configuration
    :param duration: tidal spinup duration
    :param local_fort14_filename: path to ``fort.14`` file
    :param local_fort13_filename: path to ``fort.13`` file
    :param link_mesh: whether to create a symbolic link from the original mesh to this configuration
    :param relative_paths: whether to write relative paths (instead of absolute)
    :param overwrite: whether to overwrite existing files
    :param platform: HPC platform to configure for
    :param adcirc_processors: number of processors to assign to ADCIRC execution
    :param slurm_account: Slurm account name
    :param job_duration: wall-clock duration of Slurm job
    :param partition: Slurm partition name
    :param email_type: Slurm email notification option
    :param email_address: Slurm email notification address
    """

    if not isinstance(directory, Path):
        directory = Path(directory)
    if local_fort13_filename is not None and not isinstance(local_fort13_filename, Path):
        local_fort13_filename = Path(local_fort13_filename)

    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)

    if not overwrite and configuration.files_exist(directory):
        LOGGER.warning(f'skipping existing directory "{directory}"')
        return directory

    setup_job_name = 'ADCIRC_SETUP_SPINUP'
    job_name = 'ADCIRC_COLDSTART_SPINUP'

    adcircpy_driver = configuration.adcircpy_driver

    if relative_paths:
        configuration.relative_to(directory, inplace=True)

    if 'nems' in configuration:
        nems = configuration.nemspy_modeling_system
        nems = ModelingSystem(
            nems.start_time - duration,
            nems.start_time,
            nems.interval,
            ocn=deepcopy(nems['OCN']),
            **nems.attributes,
        )
        processors = nems.processors
        model_executable = configuration['nems']['executable_path']
    else:
        nems = None
        processors = adcirc_processors
        model_executable = configuration['adcirc']['adcirc_executable_path']

    adcprep_path = configuration['adcirc']['adcprep_executable_path']
    aswip_path = configuration['adcirc']['aswip_executable_path']
    source_filename = configuration['adcirc']['source_filename']

    model_executable = update_path_relative(model_executable, relative_paths, directory)
    adcprep_path = update_path_relative(adcprep_path, relative_paths, directory)
    aswip_path = update_path_relative(aswip_path, relative_paths, directory)
    source_filename = update_path_relative(source_filename, relative_paths, directory)

    setup_script_filename = directory / 'setup.job'
    job_script_filename = directory / 'adcirc.job'

    if configuration.use_aswip:
        aswip_command = AswipCommand(path=aswip_path, nws=configuration['besttrack']['nws'])
    else:
        aswip_command = None

    setup_script = AdcircSetupJob(
        platform=platform,
        adcirc_mesh_partitions=adcirc_processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_partition=partition,
        slurm_run_name=setup_job_name,
        adcprep_path=adcprep_path,
        aswip_command=aswip_command,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{setup_job_name}.err.log',
        slurm_log_filename=f'{setup_job_name}.out.log',
        source_filename=source_filename,
    )

    job_script = AdcircRunJob(
        platform=platform,
        slurm_tasks=processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_run_name=job_name,
        executable=model_executable,
        slurm_partition=partition,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{job_name}.err.log',
        slurm_log_filename=f'{job_name}.out.log',
        source_filename=source_filename,
    )

    setup_script.write(setup_script_filename, overwrite=overwrite)
    job_script.write(job_script_filename, overwrite=overwrite)

    if 'nems' in configuration:
        LOGGER.debug(f'setting spinup to {duration}')

        nems.write(
            directory, overwrite=overwrite, include_version=True,
        )
        LOGGER.info(
            f'writing NEMS+ADCIRC tidal spinup configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    else:
        LOGGER.debug(
            f'writing ADCIRC tidal spinup configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    adcircpy_driver.write(
        directory,
        overwrite=overwrite,
        fort13=None if link_mesh else 'fort.13',
        fort14=None if link_mesh else 'fort.14',
        fort22='fort.22' if 'besttrack' in configuration else None,
        coldstart='fort.15',
        hotstart=None,
        driver=None,
    )
    if link_mesh:
        if local_fort13_filename is not None and local_fort13_filename.exists():
            create_symlink(local_fort13_filename, directory / 'fort.13', relative=True)
        create_symlink(local_fort14_filename, directory / 'fort.14', relative=True)

    return directory


def write_run_directory(
    directory: PathLike,
    name: str,
    phase: str,
    configuration: Union[ADCIRCRunConfiguration, NEMSADCIRCRunConfiguration],
    local_fort14_filename: PathLike,
    local_fort13_filename: PathLike = None,
    link_mesh: bool = False,
    relative_paths: bool = False,
    overwrite: bool = False,
    platform: Platform = None,
    adcirc_processors: int = None,
    slurm_account: str = None,
    job_duration: timedelta = None,
    partition: str = None,
    email_type: SlurmEmailType = None,
    email_address: str = None,
    do_spinup: bool = False,
    spinup_directory: PathLike = None,
) -> Path:
    """
    write an ADCIRC tidal spinup configuration to the given directory

    :param directory: path to output directory
    :param name: run name
    :param phase: ADCIRC run phase
    :param configuration: run configuration
    :param local_fort14_filename: path to ``fort.14`` file
    :param local_fort13_filename: path to ``fort.13`` file
    :param link_mesh: whether to create a symbolic link from the original mesh to this configuration
    :param relative_paths: whether to write relative paths (instead of absolute)
    :param overwrite: whether to overwrite existing files
    :param platform: HPC platform to configure for
    :param adcirc_processors: number of processors to assign to ADCIRC execution
    :param slurm_account: Slurm account name
    :param job_duration: wall-clock duration of Slurm job
    :param partition: Slurm partition name
    :param email_type: Slurm email notification option
    :param email_address: Slurm email notification address
    :param do_spinup: whether to use spinup results to hotstart run
    :param spinup_directory: path to spinup directory
    """

    if not isinstance(directory, Path):
        directory = Path(directory)
    if spinup_directory is not None and not isinstance(spinup_directory, Path):
        spinup_directory = Path(spinup_directory)
    if not isinstance(local_fort14_filename, Path):
        local_fort14_filename = Path(local_fort14_filename)
    if local_fort13_filename is not None and not isinstance(local_fort13_filename, Path):
        local_fort13_filename = Path(local_fort13_filename)

    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    LOGGER.debug(
        f'writing run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
    )

    if not overwrite and configuration.files_exist(directory):
        LOGGER.warning(f'skipping existing directory "{directory}"')
        return directory

    setup_job_name = f'ADCIRC_SETUP_{name}'
    job_name = f'ADCIRC_{phase}_{name}'

    adcircpy_driver = configuration.adcircpy_driver

    if relative_paths:
        configuration.relative_to(directory, inplace=True)

    if 'nems' in configuration:
        nems = configuration.nemspy_modeling_system
        processors = nems.processors
        model_executable = configuration['nems']['executable_path']
    else:
        nems = None
        processors = adcirc_processors
        model_executable = configuration['adcirc']['adcirc_executable_path']

    adcprep_path = configuration['adcirc']['adcprep_executable_path']
    aswip_path = configuration['adcirc']['aswip_executable_path']
    source_filename = configuration['adcirc']['source_filename']

    model_executable = update_path_relative(model_executable, relative_paths, directory)
    adcprep_path = update_path_relative(adcprep_path, relative_paths, directory)
    aswip_path = update_path_relative(aswip_path, relative_paths, directory)
    source_filename = update_path_relative(source_filename, relative_paths, directory)

    setup_script_filename = directory / 'setup.job'
    job_script_filename = directory / 'adcirc.job'

    if configuration.use_aswip:
        aswip_command = AswipCommand(path=aswip_path, nws=configuration['besttrack']['nws'])
    else:
        aswip_command = None

    setup_script = AdcircSetupJob(
        platform=platform,
        adcirc_mesh_partitions=adcirc_processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_partition=partition,
        slurm_run_name=setup_job_name,
        adcprep_path=adcprep_path,
        aswip_command=aswip_command,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{setup_job_name}.err.log',
        slurm_log_filename=f'{setup_job_name}.out.log',
        source_filename=source_filename,
    )

    job_script = AdcircRunJob(
        platform=platform,
        slurm_tasks=processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_run_name=job_name,
        executable=model_executable,
        slurm_partition=partition,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{job_name}.err.log',
        slurm_log_filename=f'{job_name}.out.log',
        source_filename=source_filename,
    )

    setup_script.write(setup_script_filename, overwrite=overwrite)
    job_script.write(job_script_filename, overwrite=overwrite)

    if 'nems' in configuration:
        nems.write(
            directory, overwrite=overwrite, include_version=True,
        )
        LOGGER.info(
            f'writing NEMS+ADCIRC run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    else:
        LOGGER.debug(
            f'writing ADCIRC run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    adcircpy_driver.write(
        directory,
        overwrite=overwrite,
        fort13=None if link_mesh else 'fort.13',
        fort14=None if link_mesh else 'fort.14',
        fort22='fort.22' if 'besttrack' in configuration else None,
        coldstart=None,
        hotstart='fort.15',
        driver=None,
    )

    if not link_mesh and local_fort14_filename.exists():
        link_mesh = configuration.adcircpy_mesh == AdcircMesh.open(local_fort14_filename)

    if link_mesh:
        if local_fort13_filename is not None and local_fort13_filename.exists():
            create_symlink(local_fort13_filename, directory / 'fort.13', relative=True)
        create_symlink(local_fort14_filename, directory / 'fort.14', relative=True)

    if do_spinup:
        for hotstart_filename in ['fort.67.nc', 'fort.68.nc']:
            try:
                create_symlink(
                    spinup_directory / hotstart_filename,
                    directory / hotstart_filename,
                    relative=True,
                )
            except:
                LOGGER.warning(
                    f'unable to link `{hotstart_filename}` from coldstart to hotstart; '
                    'you must manually link or copy this file after coldstart completes'
                )

    return directory


def update_path_relative(
    path: PathLike, relative: bool = False, relative_directory: PathLike = None
) -> Path:
    """
    update the given path based on the given relative path

    :param path: input path
    :param relative: whether to use a relative path to the input path
    :param relative_directory: directory to use to update input directory
    :return: updated path
    """

    if path is not None:
        if not isinstance(path, Path):
            path = Path(path)
        if relative_directory is None:
            relative_directory = Path.cwd()
        if relative:
            if not isinstance(relative_directory, Path):
                relative_directory = Path(relative_directory)
            if path.is_absolute():
                path = Path(os.path.relpath(path, relative_directory))
        elif not path.is_absolute():
            path = (relative_directory / path).resolve().absolute()
    return path



import concurrent.futures
from copy import copy, deepcopy
from datetime import datetime, timedelta
from enum import Enum
import os
from os import PathLike
from pathlib import Path
from typing import Union

from nemspy import ModelingSystem
import numpy as np

from coupledmodeldriver import Platform
from coupledmodeldriver.generate.schism.configure import (
    # NEMSSCHISMRunConfiguration,
    SCHISMRunConfiguration,
)
from coupledmodeldriver.generate.schism.script import (
    SchismEnsembleRunScript,
    SchismEnsembleCleanupScript,
    SchismRunJob,
    SchismCombineHotstartJob,
    SchismCombineSchoutJob,
)
from coupledmodeldriver.script import SlurmEmailType
from coupledmodeldriver.utilities import (
    create_symlink,
    get_logger,
    ProcessPoolExecutorStackTraced,
)

LOGGER = get_logger('cplmdldrv')


class RunPhase(Enum):
    """
    options for run phase of SCHISM execuction
    """

    COLDSTART = 'coldstart'
    HOTSTART = 'hotstart'


def generate_schism_configuration(
    configuration_directory: PathLike,
    output_directory: PathLike = None,
    relative_paths: bool = False,
    overwrite: bool = False,
    parallel: bool = True,
):
    """
    generate an SCHISM run configuration from JSON configuration files generated by ``initialize_schism``

    :param configuration_directory: path containing JSON configuration files
    :param output_directory: path to store generated configuration files
    :param relative_paths: whether to write relative paths in generated configuration files
    :param overwrite: whether to overwrite existing files
    :param parallel: generate configurations concurrently
    """

    start_time = datetime.now()

    if not isinstance(configuration_directory, Path):
        configuration_directory = Path(configuration_directory)

    if output_directory is None:
        output_directory = configuration_directory
    elif not isinstance(output_directory, Path):
        output_directory = Path(output_directory)

    output_directory = output_directory.resolve()
    if not output_directory.is_absolute():
        output_directory = output_directory.absolute()

    if not output_directory.exists():
        os.makedirs(output_directory, exist_ok=True)

    if configuration_directory.absolute().resolve() != Path.cwd():
        starting_directory = Path.cwd()
        os.chdir(configuration_directory)
        configuration_directory = Path.cwd()
    else:
        starting_directory = None

    use_nems = 'configure_nems.json' in [
        filename.name.lower() for filename in configuration_directory.iterdir()
    ]

    if use_nems:
        LOGGER.debug(f'generating NEMS configuration')
        pass
        # TODO
        # base_configuration = NEMSSCHISMRunConfiguration.read_directory(configuration_directory)
    else:
        LOGGER.debug(f'generating SCHISM-only configuration')
        base_configuration = SCHISMRunConfiguration.read_directory(configuration_directory)

    base_configuration.move_paths(configuration_directory)

    platform = base_configuration['modeldriver']['platform']

    job_duration = base_configuration['slurm']['job_duration']
    slurm_account = base_configuration['slurm']['account']
    partition = base_configuration['slurm']['partition']
    email_type = base_configuration['slurm']['email_type']
    email_address = base_configuration['slurm']['email_address']

    # TODO update functions for SCHISM
    original_fgrid_filename = base_configuration['schism']['fgrid_path']
    original_hgrid_filename = base_configuration['schism']['hgrid_path']
    schism_processors = base_configuration['schism']['processors']
    spinup_duration = base_configuration['schism']['tidal_spinup_duration']
    use_original_mesh = base_configuration['schism']['use_original_mesh']

    if original_hgrid_filename is None or not original_hgrid_filename.exists():
        raise FileNotFoundError(f'horizontal grid not found at "{original_hgrid_filename}"')

    local_fgrid_filename = output_directory / original_fgrid_filename.name
    local_hgrid_filename = output_directory / 'hgrid.gr3'
    local_param_filename = output_directory / 'param.nml'

    do_spinup = spinup_duration is not None

    run_phase = 'HOTSTART' if do_spinup else 'COLDSTART'

    if slurm_account is None:
        slurm_account = platform.value['slurm_account']

    ensemble_run_script_filename = output_directory / f'run_{platform.name.lower()}.sh'
    ensemble_cleanup_script_filename = output_directory / f'cleanup.sh'

    if use_original_mesh:
        LOGGER.info(
            f'using original mesh from "{os.path.relpath(original_hgrid_filename.resolve(), Path.cwd())}"'
        )
        if original_fgrid_filename.exists():
            create_symlink(original_fgrid_filename, local_fgrid_filename)
        create_symlink(original_hgrid_filename, local_hgrid_filename)
    elif overwrite or not local_hgrid_filename.exists():
        LOGGER.info(
            f'rewriting original mesh to "{os.path.relpath(local_hgrid_filename.resolve(), Path.cwd())}"'
        )
        try:
            # TODO: COPY MESH FILE
            ...
        except Exception as error:
            LOGGER.warning(error)

    if local_param_filename.exists():
        os.remove(local_param_filename)

    if local_fgrid_filename.exists():
        base_configuration['schism']['fgrid_path'] = local_fgrid_filename
    if local_hgrid_filename.exists():
        base_configuration['schism']['hgrid_path'] = local_hgrid_filename

    runs_directory = output_directory / 'runs'
    if not runs_directory.exists():
        runs_directory.mkdir(parents=True, exist_ok=True)

    perturbations = base_configuration.perturb()

    LOGGER.info(
        f'generating {len(perturbations)} run configuration(s) in "{os.path.relpath(runs_directory.resolve(), Path.cwd())}"'
    )

    if parallel:
        process_pool = ProcessPoolExecutorStackTraced()
        LOGGER.info(f'leveraging {os.cpu_count()} processor(s)')
    else:
        process_pool = None

    futures = []

    if do_spinup:
        spinup_directory = output_directory / 'spinup'

        spinup_configuration = copy(base_configuration)

        if not spinup_configuration.files_exist(spinup_directory):
            spinup_kwargs = {
                'directory': spinup_directory,
                'configuration': spinup_configuration,
                'duration': spinup_duration,
                'local_fgrid_filename': local_fgrid_filename,
                'local_hgrid_filename': local_hgrid_filename,
                'link_mesh': True,
                'relative_paths': relative_paths,
                'overwrite': overwrite,
                'platform': platform,
                'schism_processors': schism_processors,
                'slurm_account': slurm_account,
                'job_duration': job_duration,
                'partition': partition,
                'email_type': email_type,
                'email_address': email_address,
            }

            if parallel:
                futures.append(process_pool.submit(write_spinup_directory, **spinup_kwargs))
            else:
                spinup_directory = write_spinup_directory(**spinup_kwargs)
                LOGGER.info(f'wrote configuration to "{spinup_directory}"')
    else:
        spinup_directory = None

    for run_name, run_configuration in perturbations.items():
        run_directory = runs_directory / run_name

        if not run_configuration.files_exist(run_directory):
            run_kwargs = {
                'directory': run_directory,
                'name': run_name,
                'phase': run_phase,
                'configuration': run_configuration,
                'local_fgrid_filename': local_fgrid_filename,
                'local_hgrid_filename': local_hgrid_filename,
                'link_mesh': use_original_mesh,
                'relative_paths': relative_paths,
                'overwrite': overwrite,
                'platform': platform,
                'schism_processors': schism_processors,
                'slurm_account': slurm_account,
                'job_duration': job_duration,
                'partition': partition,
                'email_type': email_type,
                'email_address': email_address,
                'do_spinup': do_spinup,
                'spinup_directory': spinup_directory,
            }

            if parallel:
                futures.append(process_pool.submit(write_run_directory, **run_kwargs))
            else:
                write_run_directory(**run_kwargs)
                LOGGER.info(f'wrote configuration to "{run_directory}"')

    if parallel:
        for completed_future in concurrent.futures.as_completed(futures):
            LOGGER.info(f'wrote configuration to "{completed_future.result()}"')

    cleanup_script = SchismEnsembleCleanupScript()
    LOGGER.debug(
        f'writing cleanup script "{os.path.relpath(ensemble_cleanup_script_filename.resolve(), Path.cwd())}"'
    )
    cleanup_script.write(filename=ensemble_cleanup_script_filename, overwrite=overwrite)

    LOGGER.info(
        f'writing ensemble run script "{os.path.relpath(ensemble_run_script_filename.resolve(), Path.cwd())}"'
    )
    run_job_script = SchismEnsembleRunScript(
        platform=platform,
        commands=[
            'echo deleting previous SCHISM output',
            f'sh {ensemble_cleanup_script_filename.name}',
            'echo deleting previous SCHISM logs',
            'rm spinup/*.log runs/*/*.log',
        ],
        run_spinup=do_spinup,
    )
    run_job_script.write(ensemble_run_script_filename, overwrite=overwrite)

    if starting_directory is not None:
        os.chdir(starting_directory)

    LOGGER.info(f'finished in {datetime.now() - start_time}')

    if starting_directory is not None:
        os.chdir(starting_directory)

    LOGGER.info(f'finished in {datetime.now() - start_time}')


def write_spinup_directory(
    directory: PathLike,
    # configuration: Union[SCHISMRunConfiguration, NEMSSCHISMRunConfiguration],
    configuration: SCHISMRunConfiguration,
    duration: timedelta,
    local_hgrid_filename: PathLike,
    local_fgrid_filename: PathLike = None,
    link_mesh: bool = False,
    relative_paths: bool = False,
    overwrite: bool = False,
    platform: Platform = None,
    schism_processors: int = None,
    slurm_account: str = None,
    job_duration: timedelta = None,
    partition: str = None,
    email_type: SlurmEmailType = None,
    email_address: str = None,
) -> Path:
    """
    write an SCHISM tidal spinup configuration to the given directory

    :param directory: path to output directory
    :param configuration: run configuration
    :param duration: tidal spinup duration
    :param relative_paths: whether to write relative paths (instead of absolute)
    :param overwrite: whether to overwrite existing files
    :param platform: HPC platform to configure for
    :param schism_processors: number of processors to assign to SCHISM execution
    :param slurm_account: Slurm account name
    :param job_duration: wall-clock duration of Slurm job
    :param partition: Slurm partition name
    :param email_type: Slurm email notification option
    :param email_address: Slurm email notification address
    """

    if not isinstance(directory, Path):
        directory = Path(directory)

    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)

    if not overwrite and configuration.files_exist(directory):
        LOGGER.warning(f'skipping existing directory "{directory}"')
        return directory

    job_name = 'SCHISM_COLDSTART_SPINUP'
    combine_job_name = 'SCHISM_COMBINE_HOTSTARTFILES_SPINUP'

    pyschism_driver = configuration.pyschism_driver

    start_date = configuration['schism']['modeled_start_time']
    end_date = configuration['schism']['modeled_end_time']
    spinup_duration = configuration['schism']['tidal_spinup_duration']

    pyschism_driver.param.opt.start_date = start_date - spinup_duration
    pyschism_driver.param.core.rnday = spinup_duration
    pyschism_driver.param.core.ihfskip = spinup_duration
    pyschism_driver.param.schout.nhot = 1
    pyschism_driver.param.schout.nhot_write = pyschism_driver.param.core.ihfskip

    if relative_paths:
        configuration.relative_to(directory, inplace=True)

    schism_old_io = False
    if 'nems' in configuration:
        # TODO: Check if nems uses the new or old io run
        nems = configuration.nemspy_modeling_system
        nems = ModelingSystem(
            nems.start_time - duration,
            nems.start_time,
            nems.interval,
            ocn=deepcopy(nems['OCN']),
            **nems.attributes,
        )
        processors = nems.processors
        model_executable = configuration['nems']['executable_path']
        hotstart_combiner = configuration['schism']['schism_hotstart_combiner_path']
    else:
        nems = None
        processors = schism_processors
        model_executable = configuration['schism']['schism_executable_path']
        schism_old_io = configuration['schism']['schism_use_old_io']
        hotstart_combiner = configuration['schism']['schism_hotstart_combiner_path']

    source_filename = configuration['schism']['source_filename']

    model_executable = update_path_relative(model_executable, relative_paths, directory)
    hotstart_combiner = update_path_relative(
        hotstart_combiner, relative_paths, directory / 'outputs'
    )
    source_filename = update_path_relative(source_filename, relative_paths, directory)

    job_script_filename = directory / 'schism.job'
    job_script = SchismRunJob(
        platform=platform,
        old_io=schism_old_io,
        slurm_tasks=processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_run_name=job_name,
        executable=model_executable,
        slurm_partition=partition,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{job_name}.err.log',
        slurm_log_filename=f'{job_name}.out.log',
        source_filename=source_filename,
    )
    job_script.write(job_script_filename, overwrite=overwrite)

    combine_script_filename = directory / 'combine_hotstart.job'
    combine_script = SchismCombineHotstartJob(
        platform=platform,
        iterations_idx=[pyschism_driver.param.schout.nhot_write],
        slurm_tasks=processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_run_name=combine_job_name,
        executable=hotstart_combiner,
        slurm_partition=partition,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{combine_job_name}.err.log',
        slurm_log_filename=f'{combine_job_name}.out.log',
        source_filename=source_filename,
    )
    combine_script.write(combine_script_filename, overwrite=overwrite)

    # Link to the combined hotstart file
    try:
        combined_hotstart_filename = (
            f'hotstart_it={pyschism_driver.param.schout.nhot_write}.nc'
        )
        create_symlink(
            directory / 'outputs' / combined_hotstart_filename,
            directory / 'hotstart_out.nc',
            relative=True,
        )
    except:
        LOGGER.warning(f'unable to link `{combined_hotstart_filename}`')

    if 'nems' in configuration:
        LOGGER.debug(f'setting spinup to {duration}')

        nems.write(
            directory, overwrite=overwrite, include_version=True,
        )
        LOGGER.info(
            f'writing NEMS+SCHISM tidal spinup configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    else:
        LOGGER.debug(
            f'writing SCHISM tidal spinup configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )

    pyschism_driver.write(directory, overwrite=overwrite)

    return directory


def write_run_directory(
    directory: PathLike,
    name: str,
    phase: str,
    # configuration: Union[SCHISMRunConfiguration, NEMSSCHISMRunConfiguration],
    configuration: SCHISMRunConfiguration,
    local_hgrid_filename: PathLike,
    local_fgrid_filename: PathLike = None,
    link_mesh: bool = False,
    relative_paths: bool = False,
    overwrite: bool = False,
    platform: Platform = None,
    schism_processors: int = None,
    slurm_account: str = None,
    job_duration: timedelta = None,
    partition: str = None,
    email_type: SlurmEmailType = None,
    email_address: str = None,
    do_spinup: bool = False,
    spinup_directory: PathLike = None,
) -> Path:
    """
    write an SCHISM tidal spinup configuration to the given directory

    :param directory: path to output directory
    :param name: run name
    :param phase: SCHISM run phase
    :param configuration: run configuration
    :param relative_paths: whether to write relative paths (instead of absolute)
    :param overwrite: whether to overwrite existing files
    :param platform: HPC platform to configure for
    :param schism_processors: number of processors to assign to SCHISM execution
    :param slurm_account: Slurm account name
    :param job_duration: wall-clock duration of Slurm job
    :param partition: Slurm partition name
    :param email_type: Slurm email notification option
    :param email_address: Slurm email notification address
    :param do_spinup: whether to use spinup results to hotstart run
    :param spinup_directory: path to spinup directory
    """

    if not isinstance(directory, Path):
        directory = Path(directory)
    if spinup_directory is not None and not isinstance(spinup_directory, Path):
        spinup_directory = Path(spinup_directory)

    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    LOGGER.debug(
        f'writing run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
    )

    if not overwrite and configuration.files_exist(directory):
        LOGGER.warning(f'skipping existing directory "{directory}"')
        return directory

    job_name = f'SCHISM_{phase}_{name}'
    combine_job_name = f'SCHISM_COMBINE_HOTSTARTFILES_{name}'
    combine2_job_name = f'SCHISM_COMBINE_OUTPUTFILES_{name}'

    pyschism_driver = configuration.pyschism_driver

    pyschism_driver.param.opt.dramp = None
    pyschism_driver.param.opt.drampbc = None
    pyschism_driver.param.opt.dramp_ss = None
    pyschism_driver.param.opt.drampwind = None
    if do_spinup:
        pyschism_driver.param.opt.ihot = 1

    if relative_paths:
        configuration.relative_to(directory, inplace=True)

    schism_old_io = False
    if 'nems' in configuration:
        # TODO: Check if nems uses the new or old io run
        nems = configuration.nemspy_modeling_system
        processors = nems.processors
        model_executable = configuration['nems']['executable_path']
        hotstart_combiner = configuration['schism']['schism_hotstart_combiner_path']
        output_combiner = configuration['schism']['schism_schout_combiner_path']
    else:
        nems = None
        processors = schism_processors
        model_executable = configuration['schism']['schism_executable_path']
        schism_old_io = configuration['schism']['schism_use_old_io']
        hotstart_combiner = configuration['schism']['schism_hotstart_combiner_path']
        output_combiner = configuration['schism']['schism_schout_combiner_path']

    source_filename = configuration['schism']['source_filename']

    model_executable = update_path_relative(model_executable, relative_paths, directory)
    hotstart_combiner = update_path_relative(
        hotstart_combiner, relative_paths, directory / 'outputs'
    )
    output_combiner = update_path_relative(
        output_combiner, relative_paths, directory / 'outputs'
    )
    source_filename = update_path_relative(source_filename, relative_paths, directory)

    job_script_filename = directory / 'schism.job'
    job_script = SchismRunJob(
        platform=platform,
        old_io=schism_old_io,
        slurm_tasks=processors,
        slurm_account=slurm_account,
        slurm_duration=job_duration,
        slurm_run_name=job_name,
        executable=model_executable,
        slurm_partition=partition,
        slurm_email_type=email_type,
        slurm_email_address=email_address,
        slurm_error_filename=f'{job_name}.err.log',
        slurm_log_filename=f'{job_name}.out.log',
        source_filename=source_filename,
    )
    job_script.write(job_script_filename, overwrite=overwrite)

    rnday = pyschism_driver.param.core.rnday
    ihfskip = pyschism_driver.param.core.ihfskip
    nhot = pyschism_driver.param.schout.nhot
    nhot_write = pyschism_driver.param.schout.nhot_write
    dt = pyschism_driver.param.core.dt

    if nhot == 1 and nhot_write is not None:

        # num_hs_files = int(nhot_write / ihfskip)
        num_hs_files = int(round(timedelta(days=rnday).total_seconds() / (dt * nhot_write)))
        hotstart_iters_idx = [nhot_write * (mult + 1) for mult in range(num_hs_files)]

        combine_script_filename = directory / 'combine_hotstart.job'
        combine_script = SchismCombineHotstartJob(
            platform=platform,
            iterations_idx=hotstart_iters_idx,
            slurm_tasks=processors,
            slurm_account=slurm_account,
            slurm_duration=job_duration,
            slurm_run_name=combine_job_name,
            executable=hotstart_combiner,
            slurm_partition=partition,
            slurm_email_type=email_type,
            slurm_email_address=email_address,
            slurm_error_filename=f'{combine_job_name}.err.log',
            slurm_log_filename=f'{combine_job_name}.out.log',
            source_filename=source_filename,
        )
        combine_script.write(combine_script_filename, overwrite=overwrite)

    if schism_old_io:
        end_output_stack_idx = int(
            np.ceil(np.ceil(timedelta(days=rnday).total_seconds() / dt) / ihfskip)
        )

        combine2_script_filename = directory / 'combine_output.job'
        combine2_script = SchismCombineSchoutJob(
            platform=platform,
            start_idx=1,
            end_idx=end_output_stack_idx,
            slurm_tasks=processors,
            slurm_account=slurm_account,
            slurm_duration=job_duration,
            slurm_run_name=combine2_job_name,
            executable=output_combiner,
            slurm_partition=partition,
            slurm_email_type=email_type,
            slurm_email_address=email_address,
            slurm_error_filename=f'{combine2_job_name}.err.log',
            slurm_log_filename=f'{combine2_job_name}.out.log',
            source_filename=source_filename,
        )
        combine2_script.write(combine2_script_filename, overwrite=overwrite)

    if 'nems' in configuration:
        nems.write(
            directory, overwrite=overwrite, include_version=True,
        )
        LOGGER.info(
            f'writing NEMS+SCHISM run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )
    else:
        LOGGER.debug(
            f'writing SCHISM run configuration to "{os.path.relpath(directory.resolve(), Path.cwd())}"'
        )

    pyschism_driver.write(
        directory, overwrite=overwrite,
    )

    if do_spinup:
        try:
            create_symlink(
                spinup_directory / 'hotstart_out.nc', directory / 'hotstart.nc', relative=True,
            )
        except:
            LOGGER.warning(
                f'unable to link hotstart from coldstart to hotstart; '
                'you must manually link or copy this file after coldstart completes'
            )

    return directory


def update_path_relative(
    path: PathLike, relative: bool = False, relative_directory: PathLike = None
) -> Path:
    """
    update the given path based on the given relative path

    :param path: input path
    :param relative: whether to use a relative path to the input path
    :param relative_directory: directory to use to update input directory
    :return: updated path
    """

    if path is not None:
        if not isinstance(path, Path):
            path = Path(path)
        if relative_directory is None:
            relative_directory = Path.cwd()
        if relative:
            if not isinstance(relative_directory, Path):
                relative_directory = Path(relative_directory)
            if path.is_absolute():
                path = Path(os.path.relpath(path, relative_directory))
        elif not path.is_absolute():
            path = (relative_directory / path).resolve().absolute()
    return path



from abc import ABC
from enum import Enum

from coupledmodeldriver.configure.base import ConfigurationJSON


class Model(Enum):
    """
    model information -> class name
    """

    ADCIRC = 'ADCIRC'
    SCHISM = 'SCHISM'
    TidalForcing = 'Tides'
    ATMESH = 'ATMESH'
    WW3DATA = 'WW3DATA'


class ModelJSON(ConfigurationJSON, ABC):
    """
    abstraction of a model configuration

    stores model information
    """

    def __init__(self, model: Model, **kwargs):
        if not isinstance(model, Model):
            model = Model[str(model).lower()]

        ConfigurationJSON.__init__(self, **kwargs)

        self.model = model



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import IntEnum, Enum
from os import PathLike
from pathlib import Path
import sys
from typing import Any, Dict, List

from adcircpy import Tides as ADCIRCPyTides
from adcircpy.forcing.base import Forcing as ADCIRCPyForcing
from adcircpy.forcing.tides import HAMTIDE as ADCIRCPyHAMTIDE
from adcircpy.forcing.tides.tides import TidalSource as ADCIRCPyTidalSource
from adcircpy.forcing.waves.ww3 import WaveWatch3DataForcing as ADCIRCPyWaveWatch3DataForcing
from adcircpy.forcing.winds import BestTrackForcing as ADCIRCPyBestTrackForcing
from adcircpy.forcing.winds.atmesh import (
    AtmosphericMeshForcing as ADCIRCPyAtmosphericMeshForcing,
)
from adcircpy.forcing.winds.owi import OwiForcing as ADCIRCPyOwiForcing
from nemspy.model import AtmosphericForcingEntry, WaveWatch3ForcingEntry
from pandas import DataFrame
from pyschism.forcing.bctides.tides import Tides as PySCHISMTides
from pyschism.forcing.bctides.tpxo import TPXO_ELEVATION as PySCHISMTPXO_ELEV
from pyschism.forcing.bctides.tpxo import TPXO_VELOCITY as PySCHISMTPXO_VEL
from pyschism.forcing.bctides.tides import TidalDatabase as PySCHISMTidalDatabase
from pyschism.forcing.nws import BestTrackForcing as PySCHISMBestTrackForcing
from pyschism.forcing.base import ModelForcing as PySCHISMForcing
from pyschism.forcing import NWM as PySCHISMNWM
from pyschism.forcing.source_sink.nwm import NWMElementPairings

from coupledmodeldriver.configure.base import (
    AttributeJSON,
    ConfigurationJSON,
    NEMSCapJSON,
    NoRelPath,
)
from coupledmodeldriver.configure.models import Model
from coupledmodeldriver.utilities import LOGGER

ADCIRCPY_FORCINGS = {
    'Tides': 'TidalForcingJSON',
    'AtmosphericMeshForcing': 'ATMESHForcingJSON',
    'BestTrackForcing': 'BestTrackForcingJSON',
    'OWIForcing': 'OWIForcingJSON',
    'WaveWatch3DataForcing': 'WW3DATAForcingJSON',
}

ADCIRCPY_FORCING_CLASSES = (ADCIRCPyForcing, ADCIRCPyTides)

PYSCHISM_FORCINGS = {
    'Tides': 'TidalForcingJSON',
    'BestTrackForcing': 'BestTrackForcingJSON',
    'NationalWaterModel': 'NationalWaterModelFocringJSON'
    # 'NWM : ,
    # 'GFS, etc.': 'ATMESHForcingJSON',
    # 'AtmosphericMeshForcing': 'ATMESHForcingJSON',
}

PYSCHISM_FORCING_CLASSES = (PySCHISMTides, PySCHISMForcing, PySCHISMNWM)


class TidalSource(IntEnum):
    TPXO = 1
    HAMTIDE = 2


MODEL_TIDAL_SOURCE = {
    TidalSource.TPXO: {
        Model.ADCIRC: ADCIRCPyTidalSource.TPXO,
        Model.SCHISM: PySCHISMTidalDatabase.TPXO,
    },
    TidalSource.HAMTIDE: {
        Model.ADCIRC: ADCIRCPyTidalSource.HAMTIDE,
        Model.SCHISM: PySCHISMTidalDatabase.HAMTIDE,
    },
}


class ForcingJSON(ConfigurationJSON, ABC):
    """
    abstraction of a forcing configuration
    """

    @property
    @abstractmethod
    def adcircpy_forcing(self) -> ADCIRCPyForcing:
        """
        create an ADCIRCpy forcing object with values from this configuration
        """

        raise NotImplementedError

    def to_adcircpy(self) -> ADCIRCPyForcing:
        return self.adcircpy_forcing

    @classmethod
    @abstractmethod
    def from_adcircpy(cls, forcing: ADCIRCPyForcing) -> 'ForcingJSON':
        """
        read configuration values from an ADCIRCpy forcing object
        """

        forcing_class_name = forcing.__class__.__name__
        if forcing_class_name in ADCIRCPY_FORCINGS:
            configuration_class = getattr(
                sys.modules[__name__], ADCIRCPY_FORCINGS[forcing_class_name]
            )
            return configuration_class.from_adcircpy(forcing)
        else:
            raise NotImplementedError()

    @property
    @abstractmethod
    def pyschism_forcing(self) -> PySCHISMForcing:
        """
        create an pySCHISM forcing object with values from this configuration
        """

        raise NotImplementedError

    def to_pyschism(self) -> PySCHISMForcing:
        return self.pyschism_forcing

    @classmethod
    @abstractmethod
    def from_pyschism(cls, forcing: PySCHISMForcing) -> 'ForcingJSON':
        """
        read configuration values from an pySCHISM forcing object
        """

        forcing_class_name = forcing.__class__.__name__
        if forcing_class_name in PYSCHISM_FORCINGS:
            configuration_class = getattr(
                sys.modules[__name__], PYSCHISM_FORCINGS[forcing_class_name]
            )
            return configuration_class.from_pyschism(forcing)
        else:
            raise NotImplementedError()


class TimestepForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a forcing configuration with an arbitrary timestep interval
    """

    default_interval: timedelta
    field_types = {'interval': timedelta}

    def __init__(self, interval: timedelta = None, **kwargs):
        if interval is None:
            interval = self.default_interval
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(TimestepForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)

        self['interval'] = interval


class FileForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a forcing configuration tied to a forcing file on disk
    """

    field_types = {'resource': Path}

    def __init__(self, resource: PathLike, **kwargs):
        if resource is None:
            LOGGER.warning(
                f'resource path not specified for "{self.default_filename}"; '
                f'update entry before generating configuration'
            )
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(FileForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)

        self['resource'] = resource


class FileGenForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a forcing configuration tied to a file on disk which is
    used to generate the forcing
    """

    field_types = {'resource': NoRelPath}

    def __init__(self, resource: PathLike, **kwargs):
        if resource is None:
            LOGGER.warning(
                f'resource path not specified for "{self.default_filename}"; '
                f'update entry before generating configuration'
            )
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(FileGenForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)

        self['resource'] = resource


class TidalForcingJSON(FileGenForcingJSON):
    """
    tidal configuration in ``configure_tidal.json``

    stores tidal database and constituent information

    .. code-block:: python

        configuration = TidalForcingJSON(
            tidal_source='HAMTIDE',
            constituents='all',
        )

    """

    name = 'TIDAL'
    default_filename = f'configure_tidal.json'
    field_types = {'tidal_source': TidalSource, 'constituents': List[str]}

    def __init__(
        self,
        resource: PathLike = None,
        tidal_source: TidalSource = None,
        constituents: List[str] = None,
        **kwargs,
    ):
        if constituents is None:
            constituents = 'All'
        elif not isinstance(constituents, str):
            constituents = list(constituents)
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(TidalForcingJSON.field_types)

        FileForcingJSON.__init__(self, resource=resource, **kwargs)

        self['tidal_source'] = tidal_source
        self['constituents'] = constituents

    @property
    def adcircpy_forcing(self) -> ADCIRCPyForcing:
        tides = ADCIRCPyTides(
            tidal_source=MODEL_TIDAL_SOURCE[self['tidal_source']][Model.ADCIRC],
            resource=self['resource'],
        )

        constituents = [constituent.capitalize() for constituent in self['constituents']]

        if sorted(constituents) == sorted(
            constituent.capitalize() for constituent in tides.constituents
        ):
            constituents = ['All']
        elif sorted(constituents) == sorted(
            constituent.capitalize() for constituent in tides.major_constituents
        ):
            constituents = ['Major']

        if 'All' in constituents:
            tides.use_all()
        else:
            if 'Major' in constituents:
                tides.use_major()
                constituents.remove('Major')
            for constituent in constituents:
                if constituent not in tides.active_constituents:
                    tides.use_constituent(constituent)

        self['constituents'] = list(tides.active_constituents)
        return tides

    @classmethod
    def from_adcircpy(cls, forcing: ADCIRCPyTides) -> 'TidalForcingJSON':
        # TODO: workaround for this issue: https://github.com/JaimeCalzadaNOAA/adcircpy/pull/70#discussion_r607245713
        resource = forcing.tidal_dataset.path
        if resource == ADCIRCPyHAMTIDE.OPENDAP_URL:
            resource = None

        tidal_source = getattr(TidalSource, forcing.tidal_source.name)

        return cls(
            resource=resource,
            tidal_source=tidal_source,
            constituents=forcing.active_constituents,
        )

    @property
    def pyschism_forcing(self) -> PySCHISMTides:

        # Setup resource
        if self['tidal_source'] == TidalSource.TPXO:
            # TODO: What if h and uv files are in different locations?
            tidal_database_kwargs = dict(
                h_file=self['resource'].parent / PySCHISMTPXO_ELEV,
                u_file=self['resource'].parent / PySCHISMTPXO_VEL,
            )

        elif self['tidal_source'] == TidalSource.HAMTIDE:
            tidal_database_kwargs = {'resource': self['resource']}
        else:
            raise ValueError(f"PySCHISM doesn't support {self['tidal_source']}!")

        # Setup tidal database
        tidal_database = MODEL_TIDAL_SOURCE[self['tidal_source']][Model.SCHISM].value(
            **tidal_database_kwargs
        )

        # Setup tidal constituents
        constituents = self['constituents']

        if sorted(constituents) == sorted(
            constituent.capitalize() for constituent in PySCHISMTides.constituents
        ):
            constituents = 'all'
        elif sorted(constituents) == sorted(
            constituent.capitalize() for constituent in PySCHISMTides.major_constituents
        ):
            constituents = 'major'
        if constituents[0].lower() in ['all', 'major', 'minor']:
            constituents = list(map(str.lower, constituents))

        tides = PySCHISMTides(tidal_database=tidal_database, constituents=constituents)

        self['constituents'] = list(tides.active_constituents)
        return tides

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMTides) -> 'TidalForcingJSON':

        tidal_source = getattr(TidalSource, forcing.tidal_database.__class__.__name__)

        if tidal_source == TidalSource.TPXO:
            resource = forcing.tidal_database.h.filepath()
        elif tidal_source == TidalSource.HAMTIDE:
            # Local resource for hamtide is not yet implemented in pyschism
            resource = None
        else:
            raise ValueError(f'Invalid tidal source: {tidal_source}')

        return cls(
            resource=resource,
            tidal_source=tidal_source,
            constituents=forcing.active_constituents,
        )


class WindForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a wind forcing configuration

    stores NWS parameter
    """

    default_nws: int
    field_types = {'nws': int}

    def __init__(self, nws: int = None, **kwargs):
        if nws is None:
            nws = self.default_nws
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(WindForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)

        self['nws'] = nws


BESTTRACK_ATTRIBUTES = [
    'BLADj',
    'geofactor',
    'start_date',
    'end_date',
]


class BestTrackForcingJSON(WindForcingJSON, AttributeJSON):
    """
    storm best track configuration in ``configure_besttrack.json``

    stores storm NHC code, NWS parameter, forcing read interval, start and end dates, and optionally a path to an existing `fort.22` file

    .. code-block:: python

        configuration = BestTrackForcingJSON(nhc_code='florence2018')

        configuration = BestTrackForcingJSON.from_fort22('./fort.22')

        configuration.to_adcircpy().write('output.fort.22')

    """

    name = 'BestTrack'
    default_filename = f'configure_besttrack.json'
    default_nws = 20
    default_attributes = BESTTRACK_ATTRIBUTES
    field_types = {
        'nhc_code': str,
        'interval': timedelta,
        'start_date': datetime,
        'end_date': datetime,
        'fort22_filename': NoRelPath,
    }

    def __init__(
        self,
        nhc_code: str = None,
        nws: int = None,
        interval: timedelta = None,
        start_date: datetime = None,
        end_date: datetime = None,
        fort22_filename: PathLike = None,
        dataframe: DataFrame = None,
        attributes: Dict[str, Any] = None,
        **kwargs,
    ):
        if nhc_code is None and fort22_filename is None and dataframe is None:
            LOGGER.warning("no 'nhc_code' given")

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(BestTrackForcingJSON.field_types)

        WindForcingJSON.__init__(self, nws=nws, **kwargs)
        AttributeJSON.__init__(self, attributes=attributes, **kwargs)

        self['nhc_code'] = nhc_code
        self['interval'] = interval
        self['start_date'] = start_date
        self['end_date'] = end_date
        self['fort22_filename'] = fort22_filename

        self.__dataframe = dataframe

        if self.__dataframe is not None:
            forcing = self.adcircpy_forcing
            self['nhc_code'] = forcing.nhc_code
            self['interval'] = forcing.interval
            self['start_date'] = forcing.start_date
            self['end_date'] = forcing.end_date

    @property
    def adcircpy_forcing(self) -> ADCIRCPyBestTrackForcing:
        if self['fort22_filename'] is not None:
            forcing = ADCIRCPyBestTrackForcing.from_fort22(
                self['fort22_filename'],
                nws=self['nws'],
                interval_seconds=self['interval'],
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
            if self['nhc_code'] is not None and forcing.nhc_code != self['nhc_code']:
                try:
                    forcing.nhc_code = self['nhc_code']
                    self['nhc_code'] = forcing.nhc_code
                except ConnectionError:
                    pass
        elif self.__dataframe is not None:
            forcing = ADCIRCPyBestTrackForcing(
                storm=self.__dataframe,
                nws=self['nws'],
                interval_seconds=self['interval'],
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
        elif self['nhc_code'] is not None:
            forcing = ADCIRCPyBestTrackForcing(
                storm=self['nhc_code'],
                nws=self['nws'],
                interval_seconds=self['interval'],
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
        else:
            raise ValueError(
                f'could not create `{ADCIRCPyBestTrackForcing.__name__}` object from given information'
            )

        if self['nhc_code'] is None:
            self[
                'nhc_code'
            ] = f'{forcing.basin}{forcing.storm_number}{forcing.start_date.year}'

        for name, value in self['attributes'].items():
            if value is not None:
                try:
                    setattr(forcing, name, value)
                except:
                    LOGGER.warning(
                        f'could not set `{forcing.__class__.__name__}` attribute `{name}` to `{value}`'
                    )

        return forcing

    @classmethod
    def from_adcircpy(cls, forcing: ADCIRCPyBestTrackForcing) -> 'BestTrackForcingJSON':
        return cls(
            nhc_code=forcing.nhc_code,
            nws=forcing.NWS,
            interval=forcing.interval,
            start_date=forcing.start_date,
            end_date=forcing.end_date,
            fort22_filename=forcing.filename,
            dataframe=forcing.data,
        )

    @property
    def pyschism_forcing(self) -> PySCHISMForcing:
        if self['fort22_filename'] is not None:
            forcing = PySCHISMBestTrackForcing.from_nhc_bdeck(
                self['fort22_filename'],
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
            if self['nhc_code'] is not None and forcing.nhc_code != self['nhc_code']:
                try:
                    forcing.nhc_code = self['nhc_code']
                    self['nhc_code'] = forcing.nhc_code
                except ConnectionError:
                    pass
        elif self.__dataframe is not None:
            forcing = PySCHISMBestTrackForcing(
                storm=self.__dataframe,
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
        elif self['nhc_code'] is not None:
            forcing = PySCHISMBestTrackForcing(
                storm=self['nhc_code'],
                start_date=self['start_date'],
                end_date=self['end_date'],
            )
        else:
            raise ValueError(
                f'could not create `{PySCHISMBestTrackForcing.__name__}` object from given information'
            )

        if self['nhc_code'] is None:
            self[
                'nhc_code'
            ] = f'{forcing.basin}{forcing.storm_number}{forcing.start_date.year}'

        for name, value in self['attributes'].items():
            if value is not None:
                try:
                    setattr(forcing, name, value)
                except:
                    LOGGER.warning(
                        f'could not set `{forcing.__class__.__name__}` attribute `{name}` to `{value}`'
                    )

        return forcing

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMForcing) -> 'ForcingJSON':
        return cls(
            nhc_code=forcing.nhc_code,
            start_date=forcing.start_date,
            end_date=forcing.end_date,
            dataframe=forcing.data,
            fort22_filename=forcing.filename,
        )

    @classmethod
    def from_fort22(
        cls,
        filename: PathLike,
        nws: int = None,
        interval_seconds: int = None,
        start_date: datetime = None,
        end_date: datetime = None,
    ):
        # NOTE: The forcing object is just a temporary, so it doesn't
        # matter if it's from adcircpy or pyschism
        forcing = ADCIRCPyBestTrackForcing.from_fort22(
            filename,
            nws=nws,
            interval_seconds=interval_seconds,
            start_date=start_date,
            end_date=end_date,
        )

        return cls(
            nhc_code=forcing.nhc_code,
            nws=forcing.NWS,
            interval=forcing.interval,
            start_date=forcing.start_date,
            end_date=forcing.end_date,
            fort22_filename=filename,
        )

    from_nhc_bdeck = from_fort22

    def __copy__(self) -> 'BestTrackForcingJSON':
        instance = super().__copy__()
        instance.__class__ = self.__class__
        instance.__dataframe = self.__dataframe
        return instance


class OWIForcingJSON(WindForcingJSON, TimestepForcingJSON):
    """
    OWI forcing configuration in ``configure_owi.json``

    stores NWS parameter and forcing read interval
    """

    name = 'OWI'
    default_filename = f'configure_owi.json'
    default_nws = 12
    default_interval = timedelta(hours=1)

    def __init__(self, interval: timedelta = None, **kwargs):
        WindForcingJSON.__init__(self, nws=None, **kwargs)
        TimestepForcingJSON.__init__(self, interval=interval, **kwargs)

    @property
    def adcircpy_forcing(self) -> ADCIRCPyOwiForcing:
        return ADCIRCPyOwiForcing(interval_seconds=self['interval'] / timedelta(seconds=1))

    @classmethod
    def from_adcircpy(cls, forcing: ADCIRCPyOwiForcing) -> 'OWIForcingJSON':
        return cls(interval=timedelta(seconds=forcing.interval))

    @property
    def pyschism_forcing(self) -> PySCHISMForcing:
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMForcing) -> 'ForcingJSON':
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')


class ATMESHForcingJSON(WindForcingJSON, FileForcingJSON, TimestepForcingJSON, NEMSCapJSON):
    """
    atmospheric mesh (ATMESH) configuration in ``configure_atmesh.json``

    stores NWS parameter, forcing read interval, and optionally NEMS parameters

    .. code-block:: python

        configuration = ATMESHForcingJSON(
            resource='Wind_HWRF_SANDY_Nov2018_ExtendedSmoothT.nc',
            nws=17,
            interval=timedelta(hours=1),
        )

    """

    name = 'ATMESH'
    default_filename = f'configure_atmesh.json'
    default_nws = 17
    default_interval = timedelta(hours=1)
    default_processors = 1

    def __init__(
        self,
        resource: PathLike,
        nws: int = None,
        interval: timedelta = None,
        processors: int = None,
        nems_parameters: Dict[str, str] = None,
        **kwargs,
    ):
        WindForcingJSON.__init__(self, nws=nws, **kwargs)
        FileForcingJSON.__init__(self, resource=resource, **kwargs)
        NEMSCapJSON.__init__(
            self, processors=processors, nems_parameters=nems_parameters, **kwargs
        )
        TimestepForcingJSON.__init__(self, interval=interval, **kwargs)

    @property
    def adcircpy_forcing(self) -> ADCIRCPyForcing:
        return ADCIRCPyAtmosphericMeshForcing(
            filename=self['resource'],
            nws=self['nws'],
            interval_seconds=self['interval'] / timedelta(seconds=1),
        )

    @classmethod
    def from_adcircpy(cls, forcing: ADCIRCPyAtmosphericMeshForcing) -> 'ATMESHForcingJSON':
        return cls(resource=forcing.filename, nws=forcing.NWS, interval=forcing.interval)

    @property
    def pyschism_forcing(self) -> PySCHISMForcing:
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMForcing) -> 'ForcingJSON':
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')

    @property
    def nemspy_entry(self) -> AtmosphericForcingEntry:
        return AtmosphericForcingEntry(
            filename=self['resource'], processors=self['processors'], **self['nems_parameters']
        )


class WaveForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a wave forcing configuration

    stores NRS parameter (hundredths place of NWS parameter in ``fort.15``)
    """

    default_nrs: int
    field_types = {'nrs': int}

    def __init__(self, nrs: int = None, **kwargs):
        if nrs is None:
            nrs = self.default_nrs
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(WaveForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)

        self['nrs'] = nrs


class WW3DATAForcingJSON(WaveForcingJSON, FileForcingJSON, TimestepForcingJSON, NEMSCapJSON):
    """
    WaveWatch III output file configuration in ``configure_ww3data.json``

    stores NRS parameter, forcing read interval, and optionally NEMS parameters

    .. code-block:: python

        configuration = WW3DATAForcingJSON(
            resource='ww3.HWRF.NOV2018.2012_sxy.nc',
            nrs=5,
            interval=timedelta(hours=1),
        )

    """

    name = 'WW3DATA'
    default_filename = f'configure_ww3data.json'
    default_nrs = 5
    default_interval = timedelta(hours=1)
    default_processors = 1

    def __init__(
        self,
        resource: PathLike,
        nrs: int = None,
        interval: timedelta = None,
        processors: int = None,
        nems_parameters: Dict[str, str] = None,
        **kwargs,
    ):
        WaveForcingJSON.__init__(self, nrs=nrs, **kwargs)
        FileForcingJSON.__init__(self, resource=resource, **kwargs)
        NEMSCapJSON.__init__(
            self, processors=processors, nems_parameters=nems_parameters, **kwargs
        )
        TimestepForcingJSON.__init__(self, interval=interval, **kwargs)

    @property
    def adcircpy_forcing(self) -> ADCIRCPyForcing:
        return ADCIRCPyWaveWatch3DataForcing(
            filename=self['resource'], nrs=self['nrs'], interval_seconds=self['interval'],
        )

    @classmethod
    def from_adcircpy(cls, forcing: ADCIRCPyWaveWatch3DataForcing) -> 'WW3DATAForcingJSON':
        return cls(resource=forcing.filename, nrs=forcing.NRS, interval=forcing.interval)

    @property
    def pyschism_forcing(self) -> PySCHISMForcing:
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMForcing) -> 'ForcingJSON':
        # TODO:
        raise NotImplementedError('This forcing is not supported for SCHISM')

    @property
    def nemspy_entry(self) -> WaveWatch3ForcingEntry:
        return WaveWatch3ForcingEntry(
            filename=self['resource'], processors=self['processors'], **self['nems_parameters']
        )


class HydrologyForcingJSON(ForcingJSON, ABC):
    """
    abstraction of a hydrology forcing configuration
    """

    field_types = {}

    def __init__(self, **kwargs):
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(HydrologyForcingJSON.field_types)

        ForcingJSON.__init__(self, **kwargs)


class NationalWaterModelFocringJSON(HydrologyForcingJSON, FileForcingJSON):
    """
    National water model file configuration in ``configure_nwm.json``

    .. code-block:: python
        TODO

    """

    name = 'NWM'
    default_filename = f'configure_nwm.json'
    default_aggregation_radius = None
    defaul_cache = False
    field_types = {
        'aggregation_radius': float,
        'cache': bool,
    }

    def __init__(
        self, resource: PathLike, **kwargs,
    ):
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(NationalWaterModelFocringJSON.field_types)

        HydrologyForcingJSON.__init__(self, **kwargs)
        FileForcingJSON.__init__(self, resource=resource, **kwargs)

    @property
    def adcircpy_forcing(self) -> None:
        raise NotImplementedError('ADCIRC does NOT support NWM forcing!')

    @classmethod
    def from_adcircpy(cls, forcing: None) -> 'None':
        raise NotImplementedError('ADCIRC does NOT support NWM forcing!')

    @property
    def pyschism_forcing(self) -> PySCHISMNWM:
        return PySCHISMNWM(
            aggregation_radius=self['aggregation_radius'],
            nwm_file=self['resource'],
            cache=self['cache'],
        )

    @classmethod
    def from_pyschism(cls, forcing: PySCHISMNWM) -> 'ForcingJSON':
        return cls(
            resource=forcing.nwm_file,
            aggregation_radius=forcing.aggregation_radius,
            cache=forcing.cache,
        )



from enum import Enum
from glob import glob
from os import PathLike
from pathlib import Path
import re
from typing import Any, Dict

from file_read_backwards import FileReadBackwards


class CompletionStatus(Enum):
    """
    options for completion status
    """

    NOT_CONFIGURED = 5
    NOT_STARTED = 4
    FAILED = 3
    ERROR = 2
    RUNNING = 1
    COMPLETED = 0


def is_schism_run_directory(directory: PathLike = None) -> bool:
    """
    check if the given directory has the baseline SCHISM configuration files

    :param directory: path to directory
    :return: whether the directory is a SCHISM configuration
    """

    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    required_files = ['hgrid.gr3', 'vgrid.in', 'param.nml', 'outputs', 'bctides.in']
    for filename in required_files:
        if not (directory / filename).exists():
            return False
    else:
        return True


def check_schism_completion(
    directory: PathLike = None, verbose: bool = False
) -> Dict[str, Any]:
    """
    return the status of SCHISM execution within the given directory

    :param directory: path to directory
    :param verbose: whether to include errors and detailed status checks in output
    :return: status of SCHISM execution in JSON format
    """

    if directory is None:
        directory = Path.cwd()
    elif not isinstance(directory, Path):
        directory = Path(directory)

    status = CompletionStatus.COMPLETED
    progress = 0

    completion = {entry: {} for entry in CompletionStatus}

    if not is_schism_run_directory(directory):
        completion[CompletionStatus.NOT_CONFIGURED][
            'hgrid.gr3,vgrid.in'
        ] = f'SCHISM configuration files not found'

    if verbose or status.value < CompletionStatus.NOT_CONFIGURED.value:
        # NOTE: Slurm log files are NOT very useful for SCHISM outputs
        slurm_out_log_pattern = directory / 'SCHISM_*_*.out.log'
        slurm_output_log_filenames = [
            Path(filename) for filename in glob(str(slurm_out_log_pattern))
        ]
        if len(slurm_output_log_filenames) > 0:
            for filename in slurm_output_log_filenames:
                if verbose or status.value < CompletionStatus.ERROR.value:
                    with FileReadBackwards(filename) as log_file:
                        ended = False
                        for line in log_file:
                            if not ended and 'End Epilogue' in line:
                                ended = True

                        if not ended:
                            if filename.name not in completion[CompletionStatus.RUNNING]:
                                completion[CompletionStatus.RUNNING][filename.name] = []
                            completion[CompletionStatus.RUNNING][
                                filename.name
                            ] = f'job is still running (no `Epilogue`)'
        else:
            completion[CompletionStatus.NOT_STARTED][
                slurm_out_log_pattern.name
            ] = f'Slurm output log files `{slurm_out_log_pattern.name}` not found'

    if verbose or status.value < CompletionStatus.ERROR.value:
        slurm_error_log_pattern = directory / 'SCHISM_*_*.err.log'
        slurm_error_log_filenames = [
            Path(filename) for filename in glob(str(slurm_error_log_pattern))
        ]
        if len(slurm_error_log_filenames) > 0:
            for filename in slurm_error_log_filenames:
                if verbose or status.value < CompletionStatus.ERROR.value:
                    with open(filename) as log_file:
                        lines = list(log_file.readlines())
                        if len(lines) > 0:
                            if filename.name not in completion[CompletionStatus.ERROR]:
                                completion[CompletionStatus.ERROR][filename.name] = []
                            completion[CompletionStatus.ERROR][filename.name].extend(lines)

    if verbose or status.value < CompletionStatus.FAILED.value:
        schism_output_log_filename = directory / 'outputs' / 'mirror.out'
        if not schism_output_log_filename.exists():
            completion[CompletionStatus.FAILED][
                schism_output_log_filename.name
            ] = f'SCHISM output file `{schism_output_log_filename.name}` not found'
        else:
            stepping_init_pattern = re.compile(
                '^\stime stepping begins\.\.\.\s+\d+\s+(\d+).*$'
            )
            stepping_run_pattern = re.compile('^TIME STEP=\s+(\d+);.*$')
            finish_pattern = re.compile('^Run completed successfully.*$')
            with FileReadBackwards(schism_output_log_filename) as log_file:
                total_steps = -1
                current_step = -1
                for line in log_file:
                    match_fini = finish_pattern.match(line)
                    match_init = stepping_init_pattern.match(line)
                    match_step = stepping_run_pattern.match(line)
                    if match_fini:
                        progress = 100
                        break
                    elif match_init:
                        total_steps = int(match_init.group(1))
                    elif match_step:
                        current_step = max(current_step, int(match_step.group(1)))
                if progress != 100 and current_step > -1 and total_steps > -1:
                    progress = round(100 * current_step / total_steps, 2)

    if verbose or status.value < CompletionStatus.NOT_STARTED.value:
        esmf_log_pattern = directory / 'PET*.ESMF_LogFile'
        esmf_log_filenames = [Path(filename) for filename in glob(str(esmf_log_pattern))]
        if len(esmf_log_filenames) > 0:
            error_pattern = re.compile('error', re.IGNORECASE)
            for filename in esmf_log_filenames:
                with open(filename) as log_file:
                    lines = list(log_file.readlines())
                    if len(lines) == 0:
                        completion[CompletionStatus.FAILED][
                            filename.name
                        ] = f'empty ESMF log file `{filename.name}`'
                    else:
                        for line in lines:
                            if re.match(error_pattern, line):
                                if filename.name not in completion[CompletionStatus.ERROR]:
                                    completion[CompletionStatus.ERROR][filename.name] = []
                                completion[CompletionStatus.ERROR][filename.name].append(line)

    if verbose or status.value < CompletionStatus.NOT_STARTED.value:
        output_netcdf_pattern = directory / 'outputs' / 'schout*.nc'
        output_netcdf_files = [Path(filename) for filename in glob(str(output_netcdf_pattern))]
        if (directory / 'outputs' / 'out2d_1.nc').is_file():
            output_netcdf_files.append(directory / 'outputs' / 'out2d_1.nc')
        for filename in output_netcdf_files:
            if filename.exists():
                minimum_file_size = 43081

                if filename.stat().st_size <= minimum_file_size:
                    completion[CompletionStatus.NOT_STARTED][
                        filename.name
                    ] = f'file size ({filename.stat().st_size}) does not exceed the minimum file size ({minimum_file_size})'
            else:
                completion[CompletionStatus.NOT_STARTED][
                    filename.name
                ] = f'output file `{filename.name}` not found'

    if progress > 0 and status == CompletionStatus.NOT_STARTED:
        status = CompletionStatus.RUNNING

    for entry in CompletionStatus:
        if len(completion[entry]) > 0 and entry.value > status.value:
            status = entry

    return {
        'status': status,
        'progress': progress,
        **{entry.name.lower(): value for entry, value in completion.items()},
    }



from coupledmodeldriver.generate.adcirc.configure import (
    ADCIRCRunConfiguration,
    NEMSADCIRCRunConfiguration,
)
from coupledmodeldriver.generate.adcirc.generate import generate_adcirc_configuration



from copy import copy
from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List

from adcircpy import AdcircMesh, AdcircRun
from nemspy import ModelingSystem
from nemspy.model.base import ModelEntry

from coupledmodeldriver.configure import NEMSJSON
from coupledmodeldriver.configure.base import (
    ConfigurationJSON,
    ModelDriverJSON,
    NEMSCapJSON,
    SlurmJSON,
)
from coupledmodeldriver.configure.configure import RunConfiguration
from coupledmodeldriver.configure.forcings.base import (
    ATMESHForcingJSON,
    BestTrackForcingJSON,
    ForcingJSON,
    OWIForcingJSON,
    TidalForcingJSON,
    WW3DATAForcingJSON,
)
from coupledmodeldriver.generate.adcirc.base import ADCIRCJSON
from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.utilities import LOGGER


class ADCIRCRunConfiguration(RunConfiguration):
    """
    run configuration for ADCIRC-only run, with optional tidal and / or best track forcing using ADCIRC's input capability
    """

    REQUIRED = {
        ModelDriverJSON,
        SlurmJSON,
        ADCIRCJSON,
    }
    SUPPLEMENTARY = {
        TidalForcingJSON,
        BestTrackForcingJSON,
        OWIForcingJSON,
        ATMESHForcingJSON,
        WW3DATAForcingJSON,
    }

    def __init__(
        self,
        mesh_directory: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        modeled_timestep: timedelta,
        tidal_spinup_duration: timedelta = None,
        platform: Platform = None,
        perturbations: Dict[str, Dict[str, Any]] = None,
        forcings: List[ForcingJSON] = None,
        adcirc_processors: int = None,
        slurm_job_duration: timedelta = None,
        slurm_partition: str = None,
        slurm_email_address: str = None,
        adcirc_executable: PathLike = None,
        adcprep_executable: PathLike = None,
        aswip_executable: PathLike = None,
        source_filename: PathLike = None,
    ):
        """
        :param mesh_directory: path to input mesh directory (containing ``fort.13``, ``fort.14``)
        :param modeled_start_time: start time within the modeled system
        :param modeled_end_time: end time within the modeled system
        :param modeled_timestep: time interval within the modeled system
        :param tidal_spinup_duration: spinup time for ADCIRC tidal coldstart
        :param platform: HPC platform for which to configure
        :param perturbations: dictionary of runs encompassing run names to parameter values
        :param forcings: list of forcing configurations to connect to ADCIRC
        :param adcirc_processors: numbers of processors to assign for ADCIRC
        :param slurm_job_duration: wall clock time of job
        :param slurm_partition: Slurm partition
        :param slurm_email_address: email address to send Slurm notifications
        :param adcirc_executable: filename of compiled ``adcirc``
        :param adcprep_executable: filename of compiled ``adcprep``
        :param aswip_executable: filename of compiled ``aswip``
        :param source_filename: path to module file to ``source``
        """

        self.__adcircpy_mesh = None

        if not isinstance(mesh_directory, Path):
            mesh_directory = Path(mesh_directory)

        if platform is None:
            platform = Platform.LOCAL

        if forcings is None:
            forcings = []

        if adcirc_processors is None:
            adcirc_processors = 11

        if adcirc_executable is None:
            adcirc_executable = 'adcirc'

        if adcprep_executable is None:
            adcprep_executable = 'adcprep'

        slurm = SlurmJSON(
            account=platform.value['slurm_account'],
            tasks=adcirc_processors,
            partition=slurm_partition,
            job_duration=slurm_job_duration,
            email_address=slurm_email_address,
        )

        adcirc = ADCIRCJSON(
            adcirc_executable_path=adcirc_executable,
            adcprep_executable_path=adcprep_executable,
            aswip_executable_path=aswip_executable,
            modeled_start_time=modeled_start_time,
            modeled_end_time=modeled_end_time,
            modeled_timestep=modeled_timestep,
            fort_13_path=mesh_directory / 'fort.13',
            fort_14_path=mesh_directory / 'fort.14',
            tidal_spinup_duration=tidal_spinup_duration,
            source_filename=source_filename,
            slurm_configuration=slurm,
            processors=adcirc_processors,
        )

        driver = ModelDriverJSON(platform=platform, perturbations=perturbations)
        super().__init__([driver, slurm, adcirc])

        for forcing in forcings:
            self.add_forcing(forcing)

    def add_forcing(self, forcing: ForcingJSON):
        if forcing not in self:
            forcing = self[self.add(forcing)]
            try:
                self['adcirc'].add_forcing(forcing)
            except Exception as error:
                LOGGER.error(error)

    @property
    def forcings(self) -> List[ForcingJSON]:
        return [
            configuration
            for configuration in self.configurations
            if isinstance(configuration, ForcingJSON)
        ]

    @property
    def adcircpy_mesh(self) -> AdcircMesh:
        return self['adcirc'].adcircpy_mesh

    @adcircpy_mesh.setter
    def adcircpy_mesh(self, adcircpy_mesh: AdcircMesh):
        self['adcirc'].adcircpy_mesh = adcircpy_mesh

    @property
    def adcircpy_driver(self) -> AdcircRun:
        return self['adcirc'].adcircpy_driver

    @adcircpy_driver.setter
    def adcircpy_driver(self, adcircpy_driver: AdcircRun):
        self['adcirc'].adcircpy_driver = adcircpy_driver

    @property
    def use_aswip(self) -> bool:
        if 'besttrack' in self:
            nws = self['besttrack']['nws']
            use_aswip = nws in [8, 19, 20, 21]
            if use_aswip and self['adcirc']['aswip_executable_path'] is None:
                use_aswip = False
                LOGGER.debug(
                    f'wind parameter {nws} but no `aswip` executable given; `aswip` will not be used'
                )
        else:
            use_aswip = False

        return use_aswip

    def files_exist(self, directory: PathLike) -> bool:
        if not isinstance(directory, Path):
            directory = Path(directory)

        if not directory.exists():
            return False

        files_to_write = [
            'fort.13',
            'fort.14',
            'fort.15',
            'setup.job',
            'adcirc.job',
        ]
        if 'nems' in self:
            files_to_write.extend(
                ['nems.configure', 'atm_namelist.rc', 'model_configure', 'config.rc',]
            )
        if self.use_aswip:
            files_to_write.append('fort.22')
        existing_files = [filename.name for filename in directory.iterdir()]

        return all([filename in existing_files for filename in files_to_write])

    def __copy__(self) -> 'ADCIRCRunConfiguration':
        return self.__class__.from_configurations(
            [copy(configuration) for configuration in self.configurations]
        )

    @classmethod
    def from_configurations(
        cls, configurations: List[ConfigurationJSON]
    ) -> 'ADCIRCRunConfiguration':
        required = {configuration_class: None for configuration_class in cls.REQUIRED}
        supplementary = {
            configuration_class: None for configuration_class in cls.SUPPLEMENTARY
        }

        for configuration in configurations:
            for configuration_class in required:
                if isinstance(configuration, configuration_class):
                    if required[configuration_class] is None:
                        required[configuration_class] = configuration
                        break
                    else:
                        raise ValueError(
                            f'multiple configurations given for "{configuration_class.__name__}"'
                        )
            for configuration_class in supplementary:
                if isinstance(configuration, configuration_class):
                    supplementary[configuration_class] = configuration

        instance = RunConfiguration(required.values())
        instance.__class__ = cls

        instance['modeldriver'] = required[ModelDriverJSON]
        instance['slurm'] = required[SlurmJSON]
        instance['adcirc'] = required[ADCIRCJSON]

        forcings = [
            configuration
            for configuration in supplementary.values()
            if isinstance(configuration, ForcingJSON)
        ]
        for forcing in forcings:
            instance.add_forcing(forcing)

        return instance

    @classmethod
    def read_directory(
        cls, directory: PathLike, required: List[type] = None, supplementary: List[type] = None
    ) -> 'ADCIRCRunConfiguration':
        if not isinstance(directory, Path):
            directory = Path(directory)
        if directory.is_file():
            directory = directory.parent
        if required is None:
            required = set()
        required.update(ADCIRCRunConfiguration.REQUIRED)
        if supplementary is None:
            supplementary = set()
        supplementary.update(ADCIRCRunConfiguration.SUPPLEMENTARY)

        return super().read_directory(directory, required, supplementary)


class NEMSADCIRCRunConfiguration(ADCIRCRunConfiguration):
    """
    run configuration coupling ADCIRC with other models / forcings using NUOPC NEMS
    """

    REQUIRED = {
        ModelDriverJSON,
        NEMSJSON,
        SlurmJSON,
        ADCIRCJSON,
    }

    def __init__(
        self,
        mesh_directory: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        modeled_timestep: timedelta,
        nems_interval: timedelta,
        nems_connections: List[str],
        nems_mediations: List[str],
        nems_sequence: List[str],
        tidal_spinup_duration: timedelta = None,
        platform: Platform = None,
        perturbations: Dict[str, Dict[str, Any]] = None,
        forcings: List[ForcingJSON] = None,
        adcirc_processors: int = None,
        slurm_job_duration: timedelta = None,
        slurm_partition: str = None,
        slurm_email_address: str = None,
        nems_executable: PathLike = None,
        adcprep_executable: PathLike = None,
        aswip_executable: PathLike = None,
        source_filename: PathLike = None,
    ):
        """
        :param mesh_directory: path to input mesh directory (containing ``fort.13``, ``fort.14``)
        :param modeled_start_time: start time within the modeled system
        :param modeled_end_time: end time within the modeled system
        :param modeled_timestep: time interval within the modeled system
        :param nems_interval: modeled time interval of main NEMS loop
        :param nems_connections: list of NEMS connections as strings (i.e. ``ATM -> OCN``)
        :param nems_mediations: list of NEMS mediations, including functions
        :param nems_sequence: list of NEMS entries in sequence order
        :param adcirc_processors: numbers of processors to assign for ADCIRC
        :param platform: HPC platform for which to configure
        :param tidal_spinup_duration: spinup time for ADCIRC tidal coldstart
        :param perturbations: dictionary of runs encompassing run names to parameter values
        :param forcings: list of forcing configurations to connect to ADCIRC
        :param slurm_job_duration: wall clock time of job
        :param slurm_partition: Slurm partition
        :param slurm_email_address: email address to send Slurm notifications
        :param nems_executable: filename of compiled ``adcirc``
        :param adcprep_executable: filename of compiled ``adcprep``
        :param aswip_executable: filename of compiled ``aswip``
        :param source_filename: path to module file to ``source``
        """

        self.__nems = None

        super().__init__(
            mesh_directory=mesh_directory,
            modeled_start_time=modeled_start_time,
            modeled_end_time=modeled_end_time,
            modeled_timestep=modeled_timestep,
            tidal_spinup_duration=tidal_spinup_duration,
            platform=platform,
            perturbations=perturbations,
            forcings=None,
            adcirc_processors=adcirc_processors,
            slurm_job_duration=slurm_job_duration,
            slurm_partition=slurm_partition,
            slurm_email_address=slurm_email_address,
            adcirc_executable=nems_executable,
            adcprep_executable=adcprep_executable,
            aswip_executable=aswip_executable,
            source_filename=source_filename,
        )

        nems = NEMSJSON(
            executable_path=nems_executable,
            modeled_start_time=modeled_start_time,
            modeled_end_time=modeled_end_time,
            interval=nems_interval,
            connections=nems_connections,
            mediations=nems_mediations,
            sequence=nems_sequence,
        )

        self[nems.name.lower()] = nems

        for forcing in forcings:
            self.add_forcing(forcing)

        self['slurm'].tasks = self.nemspy_modeling_system.processors

    @property
    def nemspy_entries(self) -> List[ModelEntry]:
        return [
            configuration.nemspy_entry
            for configuration in self.configurations
            if isinstance(configuration, NEMSCapJSON)
        ]

    @property
    def nemspy_modeling_system(self) -> ModelingSystem:
        return self['nems'].to_nemspy(self.nemspy_entries)

    def add_forcing(self, forcing: ForcingJSON):
        if forcing not in self:
            forcing = self[self.add(forcing)]
            self['adcirc'].add_forcing(forcing)

    @classmethod
    def from_configurations(
        cls, configurations: List[ConfigurationJSON]
    ) -> 'NEMSADCIRCRunConfiguration':
        instance = ADCIRCRunConfiguration.from_configurations(configurations)
        instance.__class__ = cls

        nems = None
        for configuration in configurations:
            if isinstance(configuration, NEMSJSON):
                if nems is None:
                    nems = configuration
                    break
                else:
                    raise ValueError(
                        f'multiple configurations given for "{NEMSJSON.__name__}"'
                    )
        instance['nems'] = nems

        return instance

    @classmethod
    def read_directory(
        cls, directory: PathLike, required: List[type] = None, supplementary: List[type] = None
    ) -> 'NEMSADCIRCRunConfiguration':
        if not isinstance(directory, Path):
            directory = Path(directory)
        if directory.is_file():
            directory = directory.parent
        if required is None:
            required = set()
        required.update(NEMSADCIRCRunConfiguration.REQUIRED)
        if supplementary is None:
            supplementary = set()
        supplementary.update(NEMSADCIRCRunConfiguration.SUPPLEMENTARY)

        return super().read_directory(directory, required, supplementary)



from coupledmodeldriver.generate.schism.configure import (
    SCHISMRunConfiguration,
    # NEMSSCHISMRunConfiguration,
)
from coupledmodeldriver.generate.schism.generate import generate_schism_configuration



from abc import ABC, abstractmethod
from datetime import timedelta
from enum import Enum
from os import PathLike
from pathlib import Path
import textwrap
from typing import List, Sequence, Iterable
import uuid

import numpy

from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.utilities import LOGGER, make_executable


class SlurmEmailType(Enum):
    """
    option for Slurm email notification
    """

    NONE = 'NONE'
    BEGIN = 'BEGIN'
    END = 'END'
    FAIL = 'FAIL'
    REQUEUE = 'REQUEUE'
    STAGE_OUT = 'STAGE_OUT'  # burst buffer stage out and teardown completed
    ALL = 'ALL'  # equivalent to BEGIN, END, FAIL, REQUEUE, and STAGE_OUT)
    TIME_LIMIT = 'TIME_LIMIT'
    TIME_LIMIT_90 = 'TIME_LIMIT_90'  # reached 90 percent of time limit
    TIME_LIMIT_80 = 'TIME_LIMIT_80'  # reached 80 percent of time limit
    TIME_LIMIT_50 = 'TIME_LIMIT_50'  # reached 50 percent of time limit
    ARRAY_TASKS = 'ARRAY_TASKS'  # send emails for each array task


class Script:
    """
    abstraction of an executable script
    """

    shebang = '#!/bin/bash'

    def __init__(self, commands: List[str]):
        if commands is None:
            commands = []
        elif isinstance(commands, str):
            commands = [commands]
        self.commands = commands

    def __str__(self) -> str:
        lines = []

        if self.shebang is not None:
            lines.append(self.shebang)

        lines.extend(self.commands)

        return '\n'.join(lines)

    def write(self, filename: PathLike, overwrite: bool = False):
        """
        write script to file

        :param filename: path to output file
        :param overwrite: whether to overwrite existing files
        """

        if not isinstance(filename, Path):
            filename = Path(filename)

        output = f'{self}\n'
        if overwrite or not filename.exists():
            with open(filename, 'w') as file:
                file.write(output)
        else:
            LOGGER.warning(f'skipping existing file "{filename}"')


class JobScript(Script):
    """
    abstraction of a Slurm job script, to run locally or from a job manager
    """

    def __init__(
        self,
        platform: Platform,
        commands: List[str],
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        slurm_email_type: SlurmEmailType = None,
        slurm_email_address: str = None,
        slurm_error_filename: PathLike = None,
        slurm_log_filename: PathLike = None,
        slurm_nodes: int = None,
        slurm_partition: str = None,
        modules: List[PathLike] = None,
        path_prefix: str = None,
        write_slurm_directory: bool = False,
    ):
        """
        :param platform: HPC to run script on
        :param commands: shell commands to run in script
        :param slurm_run_name: Slurm run name
        :param slurm_tasks: number of total tasks for Slurm to run
        :param slurm_duration: duration to run job in job manager
        :param slurm_account: Slurm account name
        :param slurm_email_type: email type
        :param slurm_email_address: email address
        :param slurm_error_filename: file path to error log file
        :param slurm_log_filename: file path to output log file
        :param slurm_nodes: number of physical nodes to run on
        :param slurm_partition: partition to run on (stampede2 only)
        :param modules: file paths to modules to load
        :param path_prefix: file path to prepend to the PATH
        :param write_slurm_directory: explicitly add directory to Slurm header when writing file
        """

        super().__init__(commands)

        if isinstance(modules, Sequence) and len(modules) == 0:
            modules = None

        if slurm_run_name is None:
            slurm_run_name = uuid.uuid4().hex

        if slurm_tasks is None:
            slurm_tasks = platform.value['processors_per_node']

        if slurm_duration is None:
            slurm_duration = timedelta(hours=1)

        if slurm_account is None:
            slurm_account = platform.value['slurm_account']

        if slurm_partition is None:
            slurm_partition = platform.value['default_partition']

        self.platform = platform

        self.slurm_tasks = slurm_tasks
        self.slurm_account = slurm_account
        self.slurm_duration = slurm_duration

        self.__slurm_run_directory = None
        self.slurm_run_name = slurm_run_name
        self.slurm_email_type = slurm_email_type
        self.slurm_email_address = slurm_email_address

        self.slurm_error_filename = slurm_error_filename
        self.slurm_log_filename = (
            slurm_log_filename if slurm_log_filename is not None else f'{slurm_run_name}.log'
        )
        self.slurm_nodes = slurm_nodes

        self.slurm_partition = slurm_partition

        self.modules = modules
        self.path_prefix = path_prefix
        self.write_slurm_directory = write_slurm_directory

    @property
    def launcher(self) -> str:
        """
        command to start processes on target system (``srun``, ``ibrun``, etc.)
        """

        return self.platform.value['launcher']

    @property
    def slurm_tasks(self) -> int:
        return self.__slurm_tasks

    @slurm_tasks.setter
    def slurm_tasks(self, slurm_tasks: int = 1):
        if slurm_tasks is not None:
            slurm_tasks = int(slurm_tasks)
        self.__slurm_tasks = slurm_tasks

    @property
    def slurm_nodes(self) -> int:
        return self.__slurm_nodes

    @slurm_nodes.setter
    def slurm_nodes(self, slurm_nodes: int):
        if slurm_nodes is None:
            slurm_nodes = numpy.ceil(
                self.slurm_tasks / self.platform.value['processors_per_node']
            )
        if slurm_nodes is not None:
            slurm_nodes = int(slurm_nodes)
        self.__slurm_nodes = slurm_nodes

    @property
    def slurm_header(self) -> str:
        lines = []

        if self.slurm_run_name is not None:
            lines.append(f'#SBATCH -J {self.slurm_run_name}')

        if self.__slurm_run_directory is not None:
            lines.append(f'#SBATCH -D {self.__slurm_run_directory}')

        if self.slurm_account is not None:
            lines.append(f'#SBATCH -A {self.slurm_account}')
        if self.slurm_email_type not in (None, SlurmEmailType.NONE):
            lines.append(f'#SBATCH --mail-type={self.slurm_email_type.value}')
            if self.slurm_email_address is not None and len(self.slurm_email_address) > 0:
                lines.append(f'#SBATCH --mail-user={self.slurm_email_address}')
            else:
                raise ValueError('missing email address')
        if self.slurm_error_filename is not None:
            lines.append(f'#SBATCH --error={self.slurm_error_filename}')
        if self.slurm_log_filename is not None:
            lines.append(f'#SBATCH --output={self.slurm_log_filename}')

        if self.slurm_tasks is not None:
            lines.append(f'#SBATCH -n {self.slurm_tasks}')
        if self.slurm_nodes is not None:
            lines.append(f'#SBATCH -N {self.slurm_nodes}')

        if self.slurm_duration is not None:
            hours, remainder = divmod(self.slurm_duration, timedelta(hours=1))
            minutes, remainder = divmod(remainder, timedelta(minutes=1))
            seconds = round(remainder / timedelta(seconds=1))
            lines.append(f'#SBATCH --time={hours:02}:{minutes:02}:{seconds:02}')

        if self.slurm_partition is not None:
            lines.append(f'#SBATCH --partition={self.slurm_partition}')

        return '\n'.join(lines)

    def __str__(self) -> str:
        lines = []

        if self.shebang is not None:
            lines.append(self.shebang)

        if self.platform.value['uses_slurm']:
            lines.extend([self.slurm_header, '', 'set -e', ''])

        if self.modules is not None:
            modules_string = ' '.join(module for module in self.modules)
            lines.extend([f'module load {modules_string}', ''])

        if self.path_prefix is not None:
            lines.extend([f'PATH={self.path_prefix}:$PATH', ''])

        lines.extend(str(command) for command in self.commands)

        return '\n'.join(lines)

    def write(self, filename: PathLike, overwrite: bool = False):
        if not isinstance(filename, Path):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / f'{self.platform.name.lower()}.job'

        if self.write_slurm_directory:
            self.__slurm_run_directory = filename.parent

        super().write(filename, overwrite)

        if self.write_slurm_directory:
            self.__slurm_run_directory = None


class EnsembleGenerationJob(JobScript):
    """
    job script to generate the ensemble configuration
    """

    def __init__(
        self,
        platform: Platform,
        generate_command: str,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        slurm_run_name: str = 'GENERATE_CONFIGURATION',
        commands: List[str] = None,
        parallel: bool = False,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if parallel:
            generate_command = f'{generate_command} --parallel'

        self.commands.extend(
            [generate_command, 'echo "use ./run_<platform>.sh to start model"']
        )


class EnsembleRunScript(Script, ABC):
    """
    script to run the ensemble, either by running it directly or by submitting model execution to the job manager
    default filename is ``run_<platform>.sh``
    """

    def __init__(
        self, platform: Platform, run_spinup: bool = True, commands: List[str] = None,
    ):
        self.platform = platform
        self.run_spinup = run_spinup
        super().__init__(commands)

    @abstractmethod
    def _spinup_lines(self):
        pass

    @abstractmethod
    def _hotstart_lines(self):
        pass

    def __str__(self) -> str:
        lines = []

        if self.shebang is not None:
            lines.append(self.shebang)

        lines.extend(
            [
                *(str(command) for command in self.commands),
                'DIRECTORY="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)"',
                '',
            ]
        )

        if self.run_spinup:
            lines.extend(self._spinup_lines())

        hotstart_lines = [
            '# run configurations',
            bash_for_loop('for hotstart in ${DIRECTORY}/runs/*/', self._hotstart_lines()),
        ]
        lines.extend(hotstart_lines)

        if self.platform.value['uses_slurm']:
            # slurm queue output https://slurm.schedmd.com/squeue.html
            squeue_command = 'squeue -u $USER -o "%.8i %4C %4D %16E %12R %8M %j" --sort i'
            echo_squeue_command = squeue_command.replace('"', r'\"')
            lines.extend(
                [
                    '',
                    '# display job queue with dependencies',
                    squeue_command,
                    f'echo {echo_squeue_command}',
                ]
            )

        return '\n'.join(lines)

    def write(self, filename: PathLike, overwrite: bool = False):
        if not isinstance(filename, Path):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / f'run_{self.platform.name.lower()}.sh'

        super().write(filename, overwrite)

        if not filename.exists() or overwrite:
            make_executable(filename)


class EnsembleCleanupScript(Script):
    """
    script for cleaning an ensemble configuration, by deleting output and log files
    """

    def __init__(
        self,
        commands: List[str] = None,
        filenames: List[PathLike] = None,
        spinup_filenames: List[PathLike] = None,
        hotstart_filenames: List[PathLike] = None,
    ):

        self._filenames = filenames if isinstance(filenames, Iterable) else []
        self._spinup_filenames = (
            spinup_filenames if isinstance(spinup_filenames, Iterable) else []
        )
        self._hotstart_filenames = (
            hotstart_filenames if isinstance(hotstart_filenames, Iterable) else []
        )
        super().__init__(commands)

    def __str__(self):
        lines = []

        if self.shebang is not None:
            lines.append(self.shebang)

        spinup_filenames = self._filenames + self._spinup_filenames
        hotstart_filenames = self._filenames + self._hotstart_filenames

        lines.extend(
            [
                *(str(command) for command in self.commands),
                'DIRECTORY="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)"',
                '',
                '# clean spinup files',
                'rm -rf '
                + ' '.join(
                    f'${{DIRECTORY}}/spinup/{filename}' for filename in spinup_filenames
                ),
                '',
                '# clean run configurations',
                'rm -rf '
                + ' '.join(
                    f'${{DIRECTORY}}/runs/*/{filename}' for filename in hotstart_filenames
                ),
            ]
        )

        return '\n'.join(lines)

    def write(self, filename: PathLike, overwrite: bool = False):
        if not isinstance(filename, Path):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / f'cleanup.sh'

        super().write(filename, overwrite)


def bash_if_statement(
    condition: str, then: List[str], *else_then: List[List[str]], indentation: str = '    '
) -> str:
    """
    create a if statement in Bash syntax using the given condition, then statement(s), and else condition(s) / statement(s)

    :param condition: boolean condition to check
    :param then: Bash statement(s) to execute if condition is met
    :param else_then: arbitrary number of Bash statement(s) to execute if condition is not met, with optional conditions (``elif``)
    :param indentation: indentation
    :return: if statement as a string
    """

    if not isinstance(then, str) and isinstance(then, Sequence):
        then = '\n'.join(then)

    condition = str(condition).strip('if ').strip('; then')

    lines = [f'if {condition}; then', textwrap.indent(then, indentation)]

    for else_block in else_then:
        if not isinstance(else_block, str) and isinstance(else_block, Sequence):
            else_block = '\n'.join(else_block)

        currently_else = else_block.startswith('else')
        currently_elif = else_block.startswith('elif')
        currently_then = else_block.startswith('then')

        previous_line = lines[-1].strip()
        hanging_else = previous_line.startswith('else') and not previous_line.endswith(';')
        hanging_elif = previous_line.startswith('elif') and not previous_line.endswith(';')

        if currently_else or currently_elif:
            if hanging_else or hanging_elif:
                lines.remove(-1)
            if currently_else:
                lines.append(else_block)
            elif currently_elif:
                else_block.strip('elif ')
                lines.append(f'elif {else_block}')
        else:
            if not hanging_else and not hanging_elif:
                lines.append('else')
            elif hanging_elif:
                lines[-1].append(';')
                if not currently_then:
                    lines[-1].append(' then')
            lines.append(textwrap.indent(else_block, indentation))

    lines.append('fi')

    return '\n'.join(lines)


def bash_for_loop(iteration: str, do: List[str], indentation='    ') -> str:
    """
    create a for loop in Bash syntax using the given variable, iterator, and do statement(s)

    :param iteration: for loop statement, such as ``for dir in ./*``
    :param do: Bash statement(s) to execute on every loop iteration
    :param indentation: indentation
    :return: for loop as a string
    """

    if not isinstance(do, str) and isinstance(do, Sequence):
        do = '\n'.join(do)

    return '\n'.join((f'{iteration}; do', textwrap.indent(do, indentation), 'done'))


def bash_function(name: str, body: List[str], indentation: str = '    ') -> str:
    """
    create a function in Bash syntax using the given name and function statement(s)

    :param name: name of function
    :param body: Bash statement(s) making up function body
    :param indentation: indentation
    :return: function as a string
    """

    if not isinstance(body, str) and isinstance(body, Sequence):
        body = '\n'.join(body)

    return '\n'.join([f'{name}() {{', textwrap.indent(body, indentation), '}'])


def bash_if(condition, command, oneline=True):

    div = ';' if oneline else '\n'
    return f'if {condition}{div}then {command}{div}fi'


def slurm_dependencies(after_ok: List[str]):
    """
    create dependency argument for sbatch cli based on input list

    :param after_ok: list of dependencies as they should appear on bash script sbatch call
    :return: either an empty string or a dependencies argument for sbatch
    """

    dependency_list = []
    if len(after_ok) > 0:
        dependency_list.append(f'--dependency=afterok:{":".join(after_ok)}')

    return ' '.join(dependency_list)


def slurm_submit_get_id(job_file: PathLike, job_id_var: str, dependencies: str = ''):
    """
    create a script to call a job via sbatch and return the job id as a named variable in bash

    :param job_file: path to the slurm script file
    :param job_id_var: bash variable name to store the submitted slurm job id
    :param dependecies: dependency argument for sbatch command
    :return: bash script to call sbatch with optional dependencies and store job id in the specified bash variable
    """

    # NOTE: `sbatch` will only use `--dependency` if it is BEFORE the job filename
    return f"{job_id_var}=$(sbatch {dependencies} {str(job_file)} | awk '{{print $NF}}')"



from enum import Enum


class Platform(Enum):
    """
    HPC platform information
    """

    HERA = {
        'source_filename': '/scratch2/COASTAL/coastal/save/shared/repositories/ADC-WW3-NWM-NEMS/modulefiles/envmodules_intel.hera',
        'processors_per_node': 40,
        'launcher': 'srun',
        'uses_slurm': True,
        'slurm_account': 'coastal',
        'default_partition': None,
    }
    ORION = {
        'source_filename': '/work/noaa/nosofs/share/repositories/ADC-WW3-NWM-NEMS/modulefiles/envmodules_intel.orion',
        'processors_per_node': 40,
        'launcher': 'srun',
        'uses_slurm': True,
        'slurm_account': 'nosofs',
        'default_partition': 'orion',
    }
    STAMPEDE2 = {
        'source_filename': '/work/07531/zrb/stampede2/builds/ADC-WW3-NWM-NEMS/modulefiles/envmodules_intel.stampede',
        'processors_per_node': 68,
        'launcher': 'ibrun',
        'uses_slurm': True,
        'slurm_account': 'coastal',
        'default_partition': 'development',
    }
    LOCAL = {
        'source_filename': None,
        'processors_per_node': 1,
        'launcher': None,
        'uses_slurm': False,
        'slurm_account': None,
        'default_partition': None,
    }
    SLURM = {
        'source_filename': None,
        'processors_per_node': 36,
        'launcher': 'srun',
        'uses_slurm': True,
        'slurm_account': None,
        'default_partition': None,
    }



from abc import ABC, abstractmethod
from copy import copy
from datetime import datetime, timedelta
import json
import os
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import Any, Dict, List, Union

from adcircpy.server import SlurmConfig as ADCIRCPySlurmConfig
from nemspy import ModelingSystem
from nemspy.model.base import ModelEntry
from pyschism.server import SlurmConfig as PySCHISMSlurmConfig
from typepigeon import convert_to_json, convert_value

from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.script import SlurmEmailType
from coupledmodeldriver.utilities import LOGGER


class NoRelPath(type(Path())):
    """
    a helper class to indicate paths that are used only during generation
    and must NOT be converted to relative paths for solve input or run scripts
    """


class ConfigurationJSON(ABC):
    """
    abstraction of a configuration JSON, including getters and setters for values and a built-in schema of field types
    """

    name: str
    default_filename: PathLike
    field_types: Dict[str, type]

    def __init__(self, fields: Dict[str, type] = None, **configuration):
        self.field_types = {key.lower(): value for key, value in self.field_types.items()}

        if not hasattr(self, 'fields'):
            self.fields = {}

        self.fields.update({key: None for key in self.field_types if key not in self.fields})

        if fields is not None:
            fields = {key.lower(): value for key, value in fields.items()}
            self.fields.update(fields)

        if not hasattr(self, 'configuration'):
            self.configuration = {field: None for field in self.fields}

        if len(configuration) > 0:
            self.configuration.update(configuration)

    def update(self, configuration: Dict[str, Any]):
        for key, value in configuration.items():
            if key in self:
                converted_value = convert_value(value, self.fields[key])
                if self[key] != converted_value:
                    value = converted_value
                else:
                    return
            self[key] = value

    def update_from_file(self, filename: PathLike):
        with open(filename) as file:
            configuration = json.load(file)
        self.update(configuration)

    def move_paths(self, relative: PathLike):
        """
        :param relative: path to which to move Path attributes
        """

        for name, value in self.configuration.items():
            if isinstance(value, Path) and not value.is_absolute():
                self[name] = PurePosixPath(move_path(value, relative).resolve())

    def relative_to(self, path: PathLike, inplace: bool = False) -> 'ConfigurationJSON':
        instance = copy(self) if not inplace else self
        for name, value in instance.configuration.items():
            if isinstance(value, Path) and not isinstance(value, NoRelPath):
                instance[name] = PurePosixPath(os.path.relpath(value.resolve(), path))
        return instance

    def __contains__(self, key: str) -> bool:
        return key in self.configuration

    def __getitem__(self, key: str) -> Any:
        return self.configuration[key]

    def __setitem__(self, key: str, value: Any):
        if key in self.fields:
            field_type = self.fields[key]
        else:
            field_type = type(value)
            LOGGER.info(
                f'adding new configuration entry "{key}: {field_type}" to {self.name}"'
            )
        self.configuration[key] = convert_value(value, field_type)
        if key not in self.fields:
            self.fields[key] = field_type

    def __eq__(self, other: 'ConfigurationJSON') -> bool:
        return other.configuration == self.configuration

    def __repr__(self):
        configuration_string = ', '.join(
            [f'{key}={repr(value)}' for key, value in self.configuration.items()]
        )
        return f'{self.__class__.__name__}({configuration_string})'

    @classmethod
    def from_dict(cls, configuration: Dict[str, Any]) -> 'ConfigurationJSON':
        return cls(**configuration)

    def to_dict(self) -> Dict[str, Any]:
        return self.configuration

    @classmethod
    def from_string(cls, string: str) -> 'ConfigurationJSON':
        """
        read JSON string

        :param string: JSON string
        :return: configuration  object
        """

        configuration = json.loads(string)

        configuration = {
            key.lower(): convert_value(value, cls.field_types[key])
            if key in cls.field_types
            else convert_to_json(value)
            for key, value in configuration.items()
        }

        return cls(**configuration)

    def __str__(self) -> str:
        configuration = convert_to_json(self.configuration)

        if any(key != key.lower() for key in configuration):
            configuration = {key.lower(): value for key, value in configuration.items()}

        return json.dumps(configuration, indent=2)

    def __copy__(self) -> 'ConfigurationJSON':
        return self.__class__(**self.configuration)

    @classmethod
    def from_file(cls, filename: PathLike) -> 'ConfigurationJSON':
        """
        read existing JSON file

        :param filename: path to JSON file
        :return: configuration object
        """

        if not isinstance(filename, Path):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / cls.default_filename

        with open(filename) as file:
            LOGGER.debug(f'reading file "{os.path.relpath(filename.resolve(), Path.cwd())}"')
            try:
                configuration = json.load(file)
            except Exception as error:
                raise SyntaxError(f'{error} in file "{filename}"')

        configuration = {
            key.lower(): convert_value(value, cls.field_types[key])
            if key in cls.field_types
            else convert_to_json(value)
            for key, value in configuration.items()
        }

        return cls(**configuration)

    def to_file(
        self, filename: PathLike = None, absolute: bool = False, overwrite: bool = False
    ):
        """
        write configuration to file

        :param filename: path to output file
        :param absolute: whether to write absolute paths
        :param overwrite: whether to overwrite existing file
        """

        if filename is None:
            filename = self['output_directory']
        elif not isinstance(filename, Path):
            filename = Path(filename)

        if filename.is_dir():
            filename = filename / self.default_filename

        if not filename.parent.exists():
            filename.mkdir(parents=True, exist_ok=True)

        for key, value in self.configuration.items():
            if isinstance(value, Path):
                if not absolute and value.is_absolute():
                    try:
                        value = Path(os.path.relpath(value, filename.absolute().parent))
                    except:
                        pass
                self.configuration[key] = value

        configuration = convert_to_json(self.configuration)

        if any(key != key.lower() for key in configuration):
            configuration = {key.lower(): value for key, value in configuration.items()}

        if overwrite or not filename.exists():
            with open(filename.absolute(), 'w') as file:
                LOGGER.debug(
                    f'writing to file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
                )
                json.dump(configuration, file, indent=2)
        else:
            LOGGER.debug(
                f'skipping existing file "{os.path.relpath(filename.resolve(), Path.cwd())}"'
            )


class SlurmJSON(ConfigurationJSON):
    """
    Slurm configuration parameters in ``configure_slurm.json``

    stores account and partition names, wall-clock time, email notification, etc.

    .. code-block:: python

        configuration = SlurmJSON(
            account='coastal',
            tasks=602,
            job_duration=timedelta(hours=6),
        )

    """

    name = 'Slurm'
    default_filename = 'configure_slurm.json'
    field_types = {
        'account': str,
        'partition': str,
        'job_duration': timedelta,
        'run_directory': Path,
        'run_name': str,
        'email_type': SlurmEmailType,
        'email_address': str,
        'log_filename': Path,
        'modules': List[str],
        'path_prefix': Path,
        'extra_commands': List[str],
        'launcher': str,
        'nodes': int,
    }

    def __init__(
        self,
        account: str,
        tasks: int = None,
        partition: str = None,
        job_duration: timedelta = None,
        run_directory: PathLike = None,
        run_name: str = None,
        email_type: SlurmEmailType = None,
        email_address: str = None,
        log_filename: PathLike = None,
        modules: List[str] = None,
        path_prefix: Path = None,
        extra_commands: List[str] = None,
        launcher: str = None,
        nodes: int = None,
        **kwargs,
    ):
        if tasks is None:
            tasks = 1

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(SlurmJSON.field_types)

        ConfigurationJSON.__init__(self, **kwargs)

        self.tasks = tasks

        self['account'] = account
        self['partition'] = partition
        self['job_duration'] = job_duration
        self['run_directory'] = run_directory
        self['run_name'] = run_name
        self['email_type'] = email_type
        self['email_address'] = email_address
        self['log_filename'] = log_filename
        self['modules'] = modules
        self['path_prefix'] = path_prefix
        self['extra_commands'] = extra_commands
        self['launcher'] = launcher
        self['nodes'] = nodes

        if self['email_type'] is None:
            if self['email_address'] is not None:
                self['email_type'] = SlurmEmailType.ALL

    def to_adcircpy(self) -> ADCIRCPySlurmConfig:
        return ADCIRCPySlurmConfig(
            account=self['account'],
            ntasks=self.tasks,
            partition=self['partition'],
            walltime=self['job_duration'],
            filename=self['filename'] if 'filename' in self else None,
            run_directory=self['run_directory'],
            run_name=self['run_name'],
            mail_type=self['email_type'],
            mail_user=self['email_address'],
            log_filename=self['log_filename'],
            modules=self['modules'],
            path_prefix=self['path_prefix'],
            extra_commands=self['extra_commands'],
            launcher=self['launcher'],
            nodes=self['nodes'],
        )

    @classmethod
    def from_adcircpy(cls, slurm_config: ADCIRCPySlurmConfig):
        instance = cls(
            account=slurm_config._account,
            tasks=slurm_config._slurm_ntasks,
            partition=slurm_config._partition,
            job_duration=slurm_config._walltime,
            run_directory=slurm_config._run_directory,
            run_name=slurm_config._run_name,
            email_type=slurm_config._mail_type,
            email_address=slurm_config._mail_user,
            log_filename=slurm_config._log_filename,
            modules=slurm_config._modules,
            path_prefix=slurm_config._path_prefix,
            extra_commands=slurm_config._extra_commands,
            launcher=slurm_config._launcher,
            nodes=slurm_config._nodes,
        )

        instance['filename'] = slurm_config._filename

    def to_pyschism(self) -> PySCHISMSlurmConfig:
        return PySCHISMSlurmConfig(
            account=self['account'],
            ntasks=self.tasks,
            partition=self['partition'],
            walltime=self['job_duration'],
            filename=self['filename'] if 'filename' in self else None,
            run_directory=self['run_directory'],
            run_name=self['run_name'],
            mail_type=self['email_type'],
            mail_user=self['email_address'],
            log_filename=self['log_filename'],
            modules=self['modules'],
            path_prefix=self['path_prefix'],
            extra_commands=self['extra_commands'],
            launcher=self['launcher'],
            nodes=self['nodes'],
        )

    @classmethod
    def from_pyschism(cls, slurm_config: PySCHISMSlurmConfig):
        instance = cls(
            account=slurm_config._account,
            tasks=slurm_config._slurm_ntasks,
            partition=slurm_config._partition,
            job_duration=slurm_config._walltime,
            run_directory=slurm_config._run_directory,
            run_name=slurm_config._run_name,
            email_type=slurm_config._mail_type,
            email_address=slurm_config._mail_user,
            log_filename=slurm_config._log_filename,
            modules=slurm_config._modules,
            path_prefix=slurm_config._path_prefix,
            extra_commands=slurm_config._extra_commands,
            launcher=slurm_config._launcher,
            nodes=slurm_config._nodes,
        )


class ModelDriverJSON(ConfigurationJSON):
    """
    model driver configuration in ``configure_modeldriver.json``

    stores platform information and a dictionary of perturbations of other configurations per each run

    .. code-block:: python

        configuration = ModelDriverJSON(
            platform=Platform.HERA,
            perturbations=None,
        )

    """

    name = 'ModelDriver'
    default_filename = f'configure_modeldriver.json'
    field_types = {
        'platform': Platform,
        'perturbations': Dict[str, Dict[str, Dict[str, Any]]],
    }

    def __init__(
        self,
        platform: Platform,
        perturbations: Dict[str, Dict[str, Dict[str, Any]]] = None,
        **kwargs,
    ):
        """
        :param platform: platform on which to run
        :param perturbations: dictionary of runs encompassing run names to parameter values
        """

        if perturbations is None:
            perturbations = {'unperturbed': None}

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(ModelDriverJSON.field_types)

        ConfigurationJSON.__init__(self, **kwargs)

        self['platform'] = platform
        self['perturbations'] = perturbations


class AttributeJSON(ConfigurationJSON, ABC):
    """
    abstraction of a configuration with an arbitrary number of custom assignable attributes
    """

    default_attributes: List[str]
    field_types = {
        'attributes': Dict[str, Any],
    }

    def __init__(self, attributes: Dict[str, Any] = None, **kwargs):
        """
        :param attributes: attributes to store
        """

        if attributes is None:
            if self.default_attributes is not None:
                attributes = {attribute: None for attribute in self.default_attributes}
            else:
                attributes = {}

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(AttributeJSON.field_types)

        ConfigurationJSON.__init__(self, **kwargs)

        self['attributes'] = attributes


class NEMSCapJSON(ConfigurationJSON, ABC):
    """
    abtraction of an individual model configuration that implements a NEMS cap

    stores NEMS-assigned processors and NEMS cap parameters
    """

    default_processors: int
    field_types = {
        'processors': int,
        'nems_parameters': Dict[str, str],
    }

    def __init__(
        self, processors: int = None, nems_parameters: Dict[str, str] = None, **kwargs
    ):
        if processors is None:
            processors = self.default_processors
        if nems_parameters is None:
            nems_parameters = {}
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(NEMSCapJSON.field_types)

        ConfigurationJSON.__init__(self, **kwargs)

        self['processors'] = processors
        self['nems_parameters'] = nems_parameters

    @abstractmethod
    def nemspy_entry(self) -> ModelEntry:
        raise NotImplementedError()


class NEMSJSON(ConfigurationJSON):
    """
    NEMS configuration in ``configure_nems.json``

    stores NEMS executable path, modeled times / interval, connections / mediations, and the order of the run sequence

    .. code-block:: python

        model_entries = [
            AtmosphericForcingEntry('Wind_HWRF_SANDY_Nov2018_ExtendedSmoothT.nc'),
            WaveWatch3ForcingEntry('ww3.HWRF.NOV2018.2012_sxy.nc'),
            ADCIRCEntry(600),
        ]

        connections = [['ATM', 'OCN'], ['WAV', 'OCN']]
        mediations = None
        sequence = [
            'ATM -> OCN',
            'WAV -> OCN',
            'ATM',
            'WAV',
            'OCN',
        ]

        configuration = NEMSJSON(
            executable_path='NEMS.x',
            modeled_start_time=datetime(2012, 10, 22, 6),
            modeled_end_time=datetime(2012, 10, 22, 6) + timedelta(days=14.5),
            interval=timedelta(hours=1),
            connections=connections,
            mediations=mediations,
            sequence=sequence,
        )

        modeling_system = configuration.to_nemspy(model_entries)

    """

    name = 'NEMS'
    default_filename = f'configure_nems.json'
    field_types = {
        'executable_path': Path,
        'modeled_start_time': datetime,
        'modeled_end_time': datetime,
        'interval': timedelta,
        'connections': List[List[str]],
        'mediations': List[str],
        'sequence': List[str],
    }

    def __init__(
        self,
        executable_path: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        interval: timedelta = None,
        connections: List[List[str]] = None,
        mediations: List[List[str]] = None,
        sequence: List[str] = None,
        **kwargs,
    ):
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(NEMSJSON.field_types)

        ConfigurationJSON.__init__(self, **kwargs)

        self['executable_path'] = executable_path
        self['modeled_start_time'] = modeled_start_time
        self['modeled_end_time'] = modeled_end_time
        self['interval'] = interval
        self['connections'] = connections
        self['mediations'] = mediations
        self['sequence'] = sequence

    def to_nemspy(self, models: List[NEMSCapJSON]) -> ModelingSystem:
        models = [
            model.nemspy_entry if isinstance(model, NEMSCapJSON) else model for model in models
        ]
        modeling_system = ModelingSystem(
            start_time=self['modeled_start_time'],
            end_time=self['modeled_end_time'],
            interval=self['interval'],
            **{model.entry_type.value.lower(): model for model in models},
        )
        for connection in self['connections']:
            modeling_system.connect(*connection)
        for mediation in self['mediations']:
            modeling_system.mediate(*mediation)

        if len(self['sequence']) > 0:
            modeling_system.sequence = self['sequence']
        else:
            self['sequence'] = modeling_system.sequence

        return modeling_system

    @classmethod
    def from_nemspy(cls, modeling_system: ModelingSystem, executable_path: PathLike = None):
        if executable_path is None:
            executable_path = 'NEMS.x'
        return cls(
            executable_path=executable_path,
            modeled_start_time=modeling_system.start_time,
            modeled_end_time=modeling_system.end_time,
            interval=modeling_system.interval,
            models=modeling_system.models,
            connections=modeling_system.connections,
            sequence=modeling_system.sequence,
        )


def move_path(path: PathLike, move: Union[PathLike, int]) -> Path:
    if not isinstance(path, Path):
        path = Path(path)

    try:
        move = int(move)
        if move <= 0:
            move = Path('.') / ('../' * (-move))
    except:
        if not isinstance(move, Path):
            move = Path(move)

    if path.is_absolute():
        output_path = path
    elif isinstance(move, int):
        output_path = Path(*path.parts[move:])
    else:
        output_path = move / path

    return output_path
